# Copyright (C) 2020 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from adi.context_manager import context_manager
from adi.rx_tx import rx


class adis16495(rx, context_manager):
    """ADIS16495 Precision, Miniature MEMS IMU."""

    _complex_data = False
    _rx_channel_names = [
        "anglvel_x",
        "anglvel_y",
        "anglvel_z",
        "accel_x",
        "accel_y",
        "accel_z",
    ]
    _device_name = ""
    _rx_data_si_type = float

    def __init__(self, uri="", name="adis16495-3"):
        """Initialize an adis16495 object.

        parameters:
            uri: type=string
                URI of IIO context with adis16495.
            name: type=string
                Name of the device. Set to adis16495-3
                by default.
        """
        trigger_name = name + "-dev0"
        context_manager.__init__(self, uri, self._device_name)

        self._ctrl = self._ctx.find_device(name)
        self._rxadc = self._ctx.find_device(name)
        # Set default trigger
        self._trigger = self._ctx.find_device(trigger_name)
        self._rxadc._set_trigger(self._trigger)
        rx.__init__(self)
        self.rx_buffer_size = 16  # Make default buffer smaller

    # device attributes
    @property
    def sample_rate(self):
        """Sample_rate: Sample rate in samples per second."""
        return self._get_iio_dev_attr("sampling_frequency")

    @sample_rate.setter
    def sample_rate(self, value):
        self._set_iio_dev_attr_str("sampling_frequency", value)

    @property
    def current_timestamp_clock(self):
        """Current_timestamp_clock: Source clock for timestamps."""
        return self._get_iio_dev_attr("current_timestamp_clock")

    @current_timestamp_clock.setter
    def current_timestamp_clock(self, value):
        self._set_iio_dev_attr_str("current_timestamp_clock", value)

    # channel attributes
    def get_temp(self):
        """Value returned in millidegrees celsius."""
        raw = self._get_iio_attr("temp0", "raw", False)
        off = self._get_iio_attr("temp0", "offset", False)
        scale = self._get_iio_attr("temp0", "scale", False)

        return (raw + off) * scale

    temperature = property(get_temp, None)

    def __get_scaled_sensor(self, channel_name: str) -> float:
        raw = self._get_iio_attr(channel_name, "raw", False)
        scale = self._get_iio_attr(channel_name, "scale", False)

        return raw * scale

    # get anglvel_x related attributes
    def get_anglvel_x(self):
        """Value returned in radians per second."""
        return self.__get_scaled_sensor("anglvel_x")

    anglvel_x = property(get_anglvel_x, None)

    @property
    def anglvel_x_filter_low_pass_3db_frequency(self):
        return self._get_iio_attr("anglvel_x", "filter_low_pass_3db_frequency", False)

    @anglvel_x_filter_low_pass_3db_frequency.setter
    def anglvel_x_filter_low_pass_3db_frequency(self, value):
        """Value returned in HZ."""
        self._set_iio_attr("anglvel_x", "filter_low_pass_3db_frequency", False, value)

    @property
    def anglvel_x_calibbias(self):
        """User bias adjustment for the x-axis."""
        return self._get_iio_attr("anglvel_x", "calibbias", False)

    @anglvel_x_calibbias.setter
    def anglvel_x_calibbias(self, value):
        self._set_iio_attr("anglvel_x", "calibbias", False, value)

    @property
    def anglvel_x_calibscale(self):
        """User scale adjustment for the x-axis."""
        return self._get_iio_attr("anglvel_x", "calibscale", False)

    @anglvel_x_calibscale.setter
    def anglvel_x_calibscale(self, value):
        self._set_iio_attr("anglvel_x", "calibscale", False, value)

    # get anglvel_y related attributes
    def get_anglvel_y(self):
        """Value returned in radians per second."""
        return self.__get_scaled_sensor("anglvel_y")

    anglvel_y = property(get_anglvel_y, None)

    @property
    def anglvel_y_filter_low_pass_3db_frequency(self):
        """Value returned in HZ."""
        return self._get_iio_attr("anglvel_y", "filter_low_pass_3db_frequency", False)

    @anglvel_y_filter_low_pass_3db_frequency.setter
    def anglvel_y_filter_low_pass_3db_frequency(self, value):
        self._set_iio_attr("anglvel_y", "filter_low_pass_3db_frequency", False, value)

    @property
    def anglvel_y_calibbias(self):
        """User bias adjustment for the y-axis."""
        return self._get_iio_attr("anglvel_y", "calibbias", False)

    @anglvel_y_calibbias.setter
    def anglvel_y_calibbias(self, value):
        self._set_iio_attr("anglvel_y", "calibbias", False, value)

    @property
    def anglvel_y_calibscale(self):
        """User scale adjustment for the y-axis."""
        return self._get_iio_attr("anglvel_y", "calibscale", False)

    @anglvel_y_calibscale.setter
    def anglvel_y_calibscale(self, value):
        self._set_iio_attr("anglvel_y", "calibscale", False, value)

    # get anglvel_z related attributes
    def get_anglvel_z(self):
        """Value returned in radians per second."""
        return self.__get_scaled_sensor("anglvel_z")

    anglvel_z = property(get_anglvel_z, None)

    @property
    def anglvel_z_filter_low_pass_3db_frequency(self):
        """Value returned in HZ."""
        return self._get_iio_attr("anglvel_z", "filter_low_pass_3db_frequency", False)

    @anglvel_z_filter_low_pass_3db_frequency.setter
    def anglvel_z_filter_low_pass_3db_frequency(self, value):
        self._set_iio_attr("anglvel_z", "filter_low_pass_3db_frequency", False, value)

    @property
    def anglvel_z_calibbias(self):
        """User bias adjustment for the z-axis."""
        return self._get_iio_attr("anglvel_z", "calibbias", False)

    @anglvel_z_calibbias.setter
    def anglvel_z_calibbias(self, value):
        self._set_iio_attr("anglvel_z", "calibbias", False, value)

    @property
    def anglvel_z_calibscale(self):
        """User scale adjustment for the z-axis."""
        return self._get_iio_attr("anglvel_z", "calibscale", False)

    @anglvel_z_calibscale.setter
    def anglvel_z_calibscale(self, value):
        self._set_iio_attr("anglvel_z", "calibscale", False, value)

    # get accel_x related attributes
    def get_accel_x(self):
        """Value returned in m/s^2."""
        return self.__get_scaled_sensor("accel_x")

    accel_x = property(get_accel_x, None)

    @property
    def accel_x_filter_low_pass_3db_frequency(self):
        """Value returned in HZ."""
        return self._get_iio_attr("accel_x", "filter_low_pass_3db_frequency", False)

    @accel_x_filter_low_pass_3db_frequency.setter
    def accel_x_filter_low_pass_3db_frequency(self, value):
        self._set_iio_attr("accel_x", "filter_low_pass_3db_frequency", False, value)

    @property
    def accel_x_calibbias(self):
        """User bias adjustment for the x-axis."""
        return self._get_iio_attr("accel_x", "calibbias", False)

    @accel_x_calibbias.setter
    def accel_x_calibbias(self, value):
        self._set_iio_attr("accel_x", "calibbias", False, value)

    @property
    def accel_x_calibscale(self):
        """User scale adjustment for the x-axis."""
        return self._get_iio_attr("accel_x", "calibscale", False)

    @accel_x_calibscale.setter
    def accel_x_calibscale(self, value):
        self._set_iio_attr("accel_x", "calibscale", False, value)

    # get accel_y related attributes
    def get_accel_y(self):
        """Value returned in m/s^2."""
        return self.__get_scaled_sensor("accel_y")

    accel_y = property(get_accel_y, None)

    @property
    def accel_y_filter_low_pass_3db_frequency(self):
        """Value returned in HZ."""
        return self._get_iio_attr("accel_y", "filter_low_pass_3db_frequency", False)

    @accel_y_filter_low_pass_3db_frequency.setter
    def accel_y_filter_low_pass_3db_frequency(self, value):
        self._set_iio_attr("accel_y", "filter_low_pass_3db_frequency", False, value)

    @property
    def accel_y_calibbias(self):
        """User bias adjustment for the y-axis."""
        return self._get_iio_attr("accel_y", "calibbias", False)

    @accel_y_calibbias.setter
    def accel_y_calibbias(self, value):
        self._set_iio_attr("accel_y", "calibbias", False, value)

    @property
    def accel_y_calibscale(self):
        """User scale adjustment for the y-axis."""
        return self._get_iio_attr("accel_y", "calibscale", False)

    @accel_y_calibscale.setter
    def accel_y_calibscale(self, value):
        self._set_iio_attr("accel_y", "calibscale", False, value)

    # get accel_y related attributes
    def get_accel_z(self):
        """Value returned in m/s^2."""
        return self.__get_scaled_sensor("accel_z")

    accel_z = property(get_accel_z, None)

    @property
    def accel_z_filter_low_pass_3db_frequency(self):
        """Value returned in HZ."""
        return self._get_iio_attr("accel_z", "filter_low_pass_3db_frequency", False)

    @accel_z_filter_low_pass_3db_frequency.setter
    def accel_z_filter_low_pass_3db_frequency(self, value):
        self._set_iio_attr("accel_z", "filter_low_pass_3db_frequency", False, value)

    @property
    def accel_z_calibbias(self):
        """User bias adjustment for the z-axis."""
        return self._get_iio_attr("accel_z", "calibbias", False)

    @accel_z_calibbias.setter
    def accel_z_calibbias(self, value):
        self._set_iio_attr("accel_z", "calibbias", False, value)

    @property
    def accel_z_calibscale(self):
        return self._get_iio_attr("accel_z", "calibscale", False)

    @accel_z_calibscale.setter
    def accel_z_calibscale(self, value):
        """User scale adjustment for the z-axis."""
        self._set_iio_attr("accel_z", "calibscale", False, value)



# Copyright (C) 2022 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import numpy as np
from adi.attribute import attribute
from adi.context_manager import context_manager
from adi.rx_tx import rx


def _sign_extend(value, nbits):
    sign_bit = 1 << (nbits - 1)
    return (value & (sign_bit - 1)) - (value & sign_bit)


def _bitmask(nbits):
    mask = 0
    for i in range(0, nbits):
        mask = (mask << 1) | 1
    return mask


class ad4630(rx, context_manager, attribute):

    """ AD4630 is low power 24-bit precision SAR ADC """

    _complex_data = False
    _data_type = "voltage"
    _device_name = ""
    _rx_channel_names = []

    """ Default part to initialize is ad4630-24. If you don't hardware test fails"""

    def __init__(self, uri="", device_name="ad4630-24"):

        context_manager.__init__(self, uri, self._device_name)

        compatible_parts = ["ad4630-24", "ad4030-24", "ad4630-16"]

        if device_name not in compatible_parts:
            raise Exception(
                "Not a compatible device: "
                + str(device_name)
                + ". Please select from "
                + str(compatible_parts)
            )
        else:
            self._ctrl = self._ctx.find_device(device_name)
            self._rxadc = self._ctx.find_device(device_name)

        _channels = []
        self.output_bits = []
        for ch in self._ctrl.channels:
            self.output_bits.append(ch.data_format.bits)
            self._rx_channel_names.append(ch.id)
            if "differential" in ch.name:
                _channels.append((ch.id, self._diff_channel(self._ctrl, ch.id)))
                if "0" in ch.name:
                    self.chan0 = self._diff_channel(self._ctrl, ch.name)
                if "1" in ch.name:
                    self.chan1 = self._diff_channel(self._ctrl, ch.name)

        rx.__init__(self)

    def rx(self):

        if not self._rx__rxbuf:
            self._rx_init_channels()
        self._rx__rxbuf.refill()
        buff = np.frombuffer(self._rx__rxbuf.read(), dtype=np.uint32)

        data = [buff[0::2], buff[1::2]]
        temp = []
        if self._num_rx_channels != 2:
            for ch in range(0, self._num_rx_channels):
                nbits = self._ctrl.channels[ch].data_format.bits
                shift = self._ctrl.channels[ch].data_format.shift
                ch_data = np.zeros(data[int(ch / 2)].shape, dtype=np.uint32)
                for index in range(0, len(data[int(ch / 2)])):
                    ch_data[index] = (data[int(ch / 2)][index] >> shift) & _bitmask(
                        nbits
                    )
                temp.append(np.vectorize(_sign_extend)(ch_data, nbits))
            data = temp
        else:
            for idx, ch_data in enumerate(data):
                nbits = self._ctrl.channels[idx].data_format.bits
                temp.append(np.vectorize(_sign_extend)(ch_data, nbits))
            data = np.vectorize(_sign_extend)(data, nbits)

        return data

    def output_data_mode(self):
        """Determine the output data mode in which device is configured."""
        if self.output_bits[0] == 30:
            return "30bit_avg"
        if self.output_bits[0] == 32:
            return "32bit_test_pattern"
        if self.output_bits[0] == 16:
            return "16bit_diff_8bit_cm"
        if len(self.output_bits) == 1 and self.output_bits[0] == 24:
            return "24bit_diff"
        if len(self.output_bits) == 2 and self.output_bits[0] == self.output_bits[1]:
            return "24bit_diff"
        else:
            return "24bit_diff_8bit_cm"

    @property
    def sample_rate(self):
        """Get/Set the sampling frequency."""
        return self._get_iio_dev_attr("sampling_frequency")

    @sample_rate.setter
    def sample_rate(self, rate):
        """Get/Set the sampling frequency."""
        if str(rate) in str(self.sample_rate_avail):
            self._set_iio_dev_attr("sampling_frequency", str(rate))
        else:
            raise ValueError(
                "Error: Sample rate not supported \nUse one of: "
                + str(self.sample_rate_avail)
            )

    @property
    def sample_rate_avail(self):
        """Get list of all the sampling frequency available."""
        return self._get_iio_dev_attr("sampling_frequency_available")

    @property
    def operating_mode_avail(self):
        """Get list of all the operating mode available."""
        return self._get_iio_dev_attr_str("operating_mode_available")

    @property
    def operating_mode(self):
        """Get/Set the operating mode."""
        return self._get_iio_dev_attr_str("operating_mode")

    @operating_mode.setter
    def operating_mode(self, mode):
        """Get/Set the operating mode."""
        if mode in self.operating_mode_avail:
            self._set_iio_dev_attr_str("operating_mode", mode)
        else:
            raise ValueError(
                "Error: Operating mode not supported \nUse one of: "
                + str(self.operating_mode_avail)
            )

    @property
    def sample_averaging_avail(self):
        """Get list of all the sample averaging values available. Only available in 30bit averaged mode."""
        return self._get_iio_dev_attr("sample_averaging_available")

    @property
    def sample_averaging(self):
        """Get/Set the sample averaging. Only available in 30bit averaged mode."""
        return self._get_iio_dev_attr_str("sample_averaging")

    @sample_averaging.setter
    def sample_averaging(self, n_sample):
        """Get/Set the sample averaging. Only available in 30bit averaged mode."""
        if str(self.sample_averaging) != "OFF":
            if str(n_sample) in str(self.sample_averaging_avail):
                self._set_iio_dev_attr("sample_averaging", str(n_sample))
            else:
                raise ValueError(
                    "Error: Number of avg samples not supported \nUse one of: "
                    + str(self.sample_averaging_avail)
                )
        else:
            raise Exception("Sample Averaging only available in 30bit averaged mode.")

    class _diff_channel(attribute):
        """AD4x30 differential channel."""

        def __init__(self, ctrl, channel_name):
            self.name = channel_name
            self._ctrl = ctrl

        @property
        def hw_gain(self):
            """Get/Set the hardwaregain of differential channel."""
            return self._get_iio_attr(self.name, "hardwaregain", False)

        @hw_gain.setter
        def hw_gain(self, gain):
            """Get/Set the hardwaregain of differential channel."""
            self._set_iio_attr(self.name, "hardwaregain", False, int(gain))

        @property
        def offset(self):
            """Get/Set the offset of differential channel."""
            return self._get_iio_attr(self.name, "offset", False, self._ctrl)

        @offset.setter
        def offset(self, offset):
            """Get/Set the offset of differential channel."""
            self._set_iio_attr(self.name, "offset", False, int(offset), self._ctrl)



# Copyright (C) 2021 Xiphos Systems Corp.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from contextlib import suppress

import paramiko


class sshfs:
    """Minimal sshfs replacement"""

    def __init__(self, address, username, password, sshargs=None):
        if address.startswith("ip:"):
            address = address[3:]
        self.address = address
        self.username = username
        self.password = password
        self.ssh = paramiko.SSHClient()

        self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy)
        if password is None:
            with suppress(paramiko.ssh_exception.AuthenticationException):
                self.ssh.connect(self.address, username=self.username, password=None)

            self.ssh.get_transport().auth_none(self.username)
        else:
            self.ssh.connect(
                self.address,
                username=self.username,
                password=self.password,
                look_for_keys=False,
                allow_agent=False,
            )

    def _run(self, cmd):
        (_, out, err) = self.ssh.exec_command(cmd)  # pylint: ignore=B601
        stdout = out.read().decode().strip()
        stderr = err.read().decode().strip()

        return stdout, stderr

    def isfile(self, path):
        stdout, _ = self._run(f"test -f {path}; echo $?")
        return stdout == "0"

    def listdir(self, path):
        stdout, _ = self._run(f"ls -1 {path}")
        return stdout.split()

    def gettext(self, path, *kargs, **kwargs):
        stdout, _ = self._run(f"cat {path}")
        return stdout



# Copyright (C) 2019 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from adi.ad9083 import ad9083
from adi.adf5610 import adf5610
from adi.adl5960 import adl5960
from adi.admv8818 import admv8818
from adi.adrf5720 import adrf5720
from adi.gen_mux import genmux
from adi.one_bit_adc_dac import one_bit_adc_dac


class fmcvna(adrf5720, ad9083, admv8818, genmux, adf5610, adl5960):
    """ FMCVNA Scalable 8-port Vector Network Analyzer Board """

    frontend = [0] * 8

    def __init__(self, uri):
        self.lo = adf5610(uri, device_name="adf5610")
        self.rfin_attenuator = adrf5720(uri, device_name="adrf5720-rfin")
        self.lo_attenuator = adrf5720(uri, device_name="adrf5720-lo")
        self.rfin_bpf = admv8818(uri, device_name="admv8818-rfin")
        self.lo_bpf = admv8818(uri, device_name="admv8818-lo")
        self.rfin_mux = genmux(uri, device_name="mux-rfin")
        self.lo_mux = genmux(uri, device_name="mux-doubler")

        for i in range(1, 9):
            self.frontend[i - 1] = adl5960(uri, device_name=f"adl5960-{i}")

        ad9083.__init__(self, uri)
        one_bit_adc_dac.__init__(self, uri)

        self._rxadc.set_kernel_buffers_count(1)



# Copyright (C) 2019 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from adi.context_manager import context_manager
from adi.jesd import jesd
from adi.obs import obs
from adi.rx_tx import rx_tx
from adi.sync_start import sync_start


class ad9371(rx_tx, context_manager, sync_start):
    """ AD9371 Transceiver """

    _complex_data = True
    _rx_channel_names = ["voltage0_i", "voltage0_q", "voltage1_i", "voltage1_q"]
    _tx_channel_names = ["voltage0", "voltage1", "voltage2", "voltage3"]
    _obs_channel_names = ["voltage0_i", "voltage0_q"]
    _device_name = ""

    def __init__(
        self, uri="", username="root", password="analog", disable_jesd_control=False
    ):
        """Initialize AD9371 interface class.
        Args:
            uri (str): URI of target platform with AD9371
            username (str): SSH username for target board. Required for JESD monitoring
            password (str): SSH password for target board. Required for JESD monitoring
            disable_jesd_control (bool): Disable JESD status monitoring over SSH
        """
        context_manager.__init__(self, uri, self._device_name)

        self._ctrl = self._ctx.find_device("ad9371-phy")
        self._rxadc = self._ctx.find_device("axi-ad9371-rx-hpc")
        self._rxobs = self._ctx.find_device("axi-ad9371-rx-obs-hpc")
        self._txdac = self._ctx.find_device("axi-ad9371-tx-hpc")

        if not disable_jesd_control and jesd:
            self._jesd = jesd(uri, username=username, password=password)

        rx_tx.__init__(self)

        self.obs = obs(self._ctx, self._rxobs, self._obs_channel_names)

    @property
    def ensm_mode(self):
        """ensm_mode: Enable State Machine State Allows real time control over
        the current state of the device. Options are: radio_on, radio_off"""
        return self._get_iio_dev_attr_str("ensm_mode")

    @ensm_mode.setter
    def ensm_mode(self, value):
        self._set_iio_dev_attr_str("ensm_mode", value)

    @property
    def gain_control_mode(self):
        """gain_control_mode: Mode of receive path AGC. Options are:
        automatic, hybrid, manual"""
        return self._get_iio_attr_str("voltage0", "gain_control_mode", False)

    @gain_control_mode.setter
    def gain_control_mode(self, value):
        self._set_iio_attr("voltage0", "gain_control_mode", False, value)

    @property
    def rx_quadrature_tracking_en_chan0(self):
        """Enable Quadrature tracking calibration for RX1"""
        return self._get_iio_attr("voltage0", "quadrature_tracking_en", False)

    @rx_quadrature_tracking_en_chan0.setter
    def rx_quadrature_tracking_en_chan0(self, value):
        self._set_iio_attr("voltage0", "quadrature_tracking_en", False, value)

    @property
    def rx_quadrature_tracking_en_chan1(self):
        """Enable Quadrature tracking calibration for RX2"""
        return self._get_iio_attr("voltage1", "quadrature_tracking_en", False)

    @rx_quadrature_tracking_en_chan1.setter
    def rx_quadrature_tracking_en_chan1(self, value):
        self._set_iio_attr("voltage1", "quadrature_tracking_en", False, value)

    @property
    def rx_hardwaregain_chan0(self):
        """rx_hardwaregain: Gain applied to RX path channel 0. Only applicable when
        gain_control_mode is set to 'manual'"""
        return self._get_iio_attr("voltage0", "hardwaregain", False)

    @rx_hardwaregain_chan0.setter
    def rx_hardwaregain_chan0(self, value):
        if self.gain_control_mode == "manual":
            self._set_iio_attr("voltage0", "hardwaregain", False, value)

    @property
    def rx_hardwaregain_chan1(self):
        """rx_hardwaregain: Gain applied to RX path channel 1. Only applicable when
        gain_control_mode is set to 'manual'"""
        return self._get_iio_attr("voltage1", "hardwaregain", False)

    @rx_hardwaregain_chan1.setter
    def rx_hardwaregain_chan1(self, value):
        if self.gain_control_mode == "manual":
            self._set_iio_attr("voltage1", "hardwaregain", False, value)

    @property
    def rx_temp_comp_gain_chan0(self):
        """rx_temp_comp_gain_chan0: """
        return self._get_iio_attr("voltage0", "temp_comp_gain", False)

    @rx_temp_comp_gain_chan0.setter
    def rx_temp_comp_gain_chan0(self, value):
        self._set_iio_attr("voltage0", "temp_comp_gain", False, value)

    @property
    def rx_temp_comp_gain_chan1(self):
        """rx_temp_comp_gain_chan1: """
        return self._get_iio_attr("voltage1", "temp_comp_gain", False)

    @rx_temp_comp_gain_chan1.setter
    def rx_temp_comp_gain_chan1(self, value):
        self._set_iio_attr("voltage1", "temp_comp_gain", False, value)

    @property
    def tx_quadrature_tracking_en_chan0(self):
        """Enable Quadrature tracking calibration for TX1"""
        return self._get_iio_attr("voltage0", "quadrature_tracking_en", True)

    @tx_quadrature_tracking_en_chan0.setter
    def tx_quadrature_tracking_en_chan0(self, value):
        self._set_iio_attr("voltage0", "quadrature_tracking_en", True, value)

    @property
    def tx_quadrature_tracking_en_chan1(self):
        """Enable Quadrature tracking calibration for TX2"""
        return self._get_iio_attr("voltage1", "quadrature_tracking_en", True)

    @tx_quadrature_tracking_en_chan1.setter
    def tx_quadrature_tracking_en_chan1(self, value):
        self._set_iio_attr("voltage1", "quadrature_tracking_en", True, value)

    @property
    def tx_hardwaregain_chan0(self):
        """tx_hardwaregain: Attenuation applied to TX path channel 0"""
        return self._get_iio_attr("voltage0", "hardwaregain", True)

    @tx_hardwaregain_chan0.setter
    def tx_hardwaregain_chan0(self, value):
        self._set_iio_attr("voltage0", "hardwaregain", True, value)

    @property
    def tx_hardwaregain_chan1(self):
        """tx_hardwaregain: Attenuation applied to TX path channel 1"""
        return self._get_iio_attr("voltage1", "hardwaregain", True)

    @tx_hardwaregain_chan1.setter
    def tx_hardwaregain_chan1(self, value):
        self._set_iio_attr("voltage1", "hardwaregain", True, value)

    @property
    def rx_rf_bandwidth(self):
        """rx_rf_bandwidth: Bandwidth of front-end analog filter of RX path"""
        return self._get_iio_attr("voltage0", "rf_bandwidth", False)

    @property
    def tx_rf_bandwidth(self):
        """tx_rf_bandwidth: Bandwidth of front-end analog filter of TX path"""
        return self._get_iio_attr("voltage0", "rf_bandwidth", True)

    @property
    def rx_enable_dec8(self):
        """rx_enable_dec8: Enable x8 decimation filter in RX path"""
        avail = self._get_iio_attr_str(
            "voltage0_i", "sampling_frequency_available", False, self._rxadc
        )
        avail = avail.strip().split(" ")
        val = self._get_iio_attr_str(
            "voltage0_i", "sampling_frequency", False, self._rxadc
        )
        return val == avail[1]

    @rx_enable_dec8.setter
    def rx_enable_dec8(self, value):
        avail = self._get_iio_attr_str(
            "voltage0_i", "sampling_frequency_available", False, self._rxadc
        )
        avail = sorted(avail.strip().split(" "))
        val = int(avail[1] if value else avail[0])
        self._set_iio_attr("voltage0_i", "sampling_frequency", False, val, self._rxadc)

    @property
    def tx_enable_int8(self):
        """tx_enable_int8: Enable x8 interpolation filter in TX path"""
        avail = self._get_iio_attr_str(
            "voltage0", "sampling_frequency_available", True, self._txdac
        )
        avail = avail.strip().split(" ")
        val = self._get_iio_attr_str(
            "voltage0", "sampling_frequency", True, self._txdac
        )
        return val == avail[1]

    @tx_enable_int8.setter
    def tx_enable_int8(self, value):
        avail = self._get_iio_attr_str(
            "voltage0", "sampling_frequency_available", True, self._txdac
        )
        avail = sorted(avail.strip().split(" "))
        val = int(avail[1] if value else avail[0])
        self._set_iio_attr("voltage0", "sampling_frequency", True, val, self._txdac)

    @property
    def rx_sample_rate(self):
        """rx_sample_rate: Sample rate RX path in samples per second
            This value will reflect the correct value when 8x decimator is enabled
        """
        dec = 8 if self.rx_enable_dec8 else 1
        return self._get_iio_attr("voltage0", "sampling_frequency", False) / dec

    @property
    def orx_sample_rate(self):
        """orx_sample_rate: Sample rate ORX path in samples per second
            This value will reflect the correct value when 8x decimator is enabled
        """
        dec = 8 if self.rx_enable_dec8 else 1
        return self._get_iio_attr("voltage2", "sampling_frequency", False) / dec

    @property
    def tx_sample_rate(self):
        """tx_sample_rate: Sample rate TX path in samples per second
            This value will reflect the correct value when 8x interpolator is enabled
        """
        dec = 8 if self.tx_enable_int8 else 1
        return self._get_iio_attr("voltage0", "sampling_frequency", True) / dec

    @property
    def rx_lo(self):
        """rx_lo: Carrier frequency of RX path"""
        return self._get_iio_attr("altvoltage0", "RX_LO_frequency", True)

    @rx_lo.setter
    def rx_lo(self, value):
        self._set_iio_attr("altvoltage0", "RX_LO_frequency", True, value)

    @property
    def tx_lo(self):
        """tx_lo: Carrier frequency of TX path"""
        return self._get_iio_attr("altvoltage1", "TX_LO_frequency", True)

    @tx_lo.setter
    def tx_lo(self, value):
        self._set_iio_attr("altvoltage1", "TX_LO_frequency", True, value)

    @property
    def sn_lo(self):
        """sn_lo: Carrier frequency of Sniffer/ORx path"""
        return self._get_iio_attr("altvoltage2", "RX_SN_LO_frequency", True)

    @sn_lo.setter
    def sn_lo(self, value):
        self._set_iio_attr("altvoltage2", "RX_SN_LO_frequency", True, value)

    @property
    def obs_gain_control_mode(self):
        """obs_gain_control_mode: Mode of Obs/Sniffer receive path AGC. Options are:
        automatic, hybrid, manual"""
        return self._get_iio_attr_str("voltage2", "gain_control_mode", False)

    @obs_gain_control_mode.setter
    def obs_gain_control_mode(self, value):
        self._set_iio_attr("voltage2", "gain_control_mode", False, value)

    @property
    def obs_hardwaregain(self):
        """obs_hardwaregain: Gain applied to Obs/Sniffer receive path chan0. Only applicable when
        obs_gain_control_mode is set to 'manual'"""
        return self._get_iio_attr("voltage2", "hardwaregain", False)

    @obs_hardwaregain.setter
    def obs_hardwaregain(self, value):
        if self.obs_gain_control_mode == "manual":
            self._set_iio_attr("voltage2", "hardwaregain", False, value)

    @property
    def obs_temp_comp_gain(self):
        """obs_temp_comp_gain: """
        return self._get_iio_attr("voltage2", "temp_comp_gain", False)

    @obs_temp_comp_gain.setter
    def obs_temp_comp_gain(self, value):
        self._set_iio_attr("voltage2", "temp_comp_gain", False, value)

    @property
    def obs_quadrature_tracking_en(self):
        """Enable Quadrature tracking calibration for OBS chan0"""
        return self._get_iio_attr("voltage2", "quadrature_tracking_en", False)

    @obs_quadrature_tracking_en.setter
    def obs_quadrature_tracking_en(self, value):
        self._set_iio_attr("voltage2", "quadrature_tracking_en", False, value)

    @property
    def obs_rf_port_select(self):
        """obs_rf_port_select: Observation path source. Options are:

        - OFF - SnRx path is disabled
        - ORX1_TX_LO – SnRx operates in observation mode on ORx1 with Tx LO synthesizer
        - ORX2_TX_LO – SnRx operates in observation mode on ORx2 with Tx LO synthesizer
        - INTERNALCALS – enables scheduled Tx calibrations while using SnRx path. The enableTrackingCals function needs to be called in RADIO_OFF state. It sets the calibration mask, which the scheduler will later use to schedule the desired calibrations. This command is issued in RADIO_OFF. Once the AD9371 moves to RADIO_ON state, the internal scheduler will use the enabled calibration mask to schedule calibrations whenever possible, based on the state of the transceiver. The Tx calibrations will not be scheduled until INTERNALCALS is selected and the Tx calibrations are enabled in the cal mask.
        - OBS_SNIFFER – SnRx operates in sniffer mode with latest selected Sniffer Input – for hardware pin control operation. In pin mode, the GPIO pins designated for ORX_MODE would select SNIFFER mode. Then MYKONOS_setSnifferChannel function would choose the channel.
        - ORX1_SN_LO – SnRx operates in observation mode on ORx1 with SNIFFER LO synthesizer
        - ORX2_SN_LO – SnRx operates in observation mode on ORx2 with SNIFFER LO synthesizer
        - SN_A – SnRx operates in sniffer mode on SnRxA with SNIFFER LO synthesizer
        - SN_B – SnRx operates in sniffer mode on SnRxB with SNIFFER LO synthesizer
        - SN_C – SnRx operates in sniffer mode on SnRxC with SNIFFER LO synthesizer

        """
        return self._get_iio_attr_str("voltage2", "rf_port_select", False)

    @obs_rf_port_select.setter
    def obs_rf_port_select(self, value):
        self._set_iio_attr("voltage2", "rf_port_select", False, value)

    @property
    def jesd204_statuses(self):
        return self._jesd.get_all_statuses()



# Copyright (C) 2020 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from adi.context_manager import context_manager
from adi.rx_tx import rx


class adis16507(rx, context_manager):
    """ ADIS16507 Precision, Miniature MEMS IMU """

    _complex_data = False
    _rx_channel_names = [
        "anglvel_x",
        "anglvel_y",
        "anglvel_z",
        "accel_x",
        "accel_y",
        "accel_z",
    ]
    _device_name = ""
    _rx_data_si_type = float

    def __init__(
        self, uri="", imu_dev_name="adis16507-3", trigger_name="adis16507-3-dev0"
    ):

        context_manager.__init__(self, uri, self._device_name)

        self._ctrl = self._ctx.find_device(imu_dev_name)
        self._rxadc = self._ctx.find_device(imu_dev_name)
        # Set default trigger
        self._trigger = self._ctx.find_device(trigger_name)
        self._rxadc._set_trigger(self._trigger)
        rx.__init__(self)
        self.rx_buffer_size = 16  # Make default buffer smaller

    @property
    def sample_rate(self):
        """sample_rate: Sample rate in samples per second"""
        return self._get_iio_dev_attr("sampling_frequency")

    @sample_rate.setter
    def sample_rate(self, value):
        self._set_iio_dev_attr_str("sampling_frequency", value)

    @property
    def filter_low_pass_3db_frequency(self):
        """filter_low_pass_3db_frequency: Bandwidth for the accelerometer and
        gyroscope channels"""
        return self._get_iio_dev_attr("filter_low_pass_3db_frequency")

    @filter_low_pass_3db_frequency.setter
    def filter_low_pass_3db_frequency(self, value):
        self._set_iio_dev_attr_str("filter_low_pass_3db_frequency", value)

    @property
    def current_timestamp_clock(self):
        """current_timestamp_clock: Source clock for timestamps"""
        return self._get_iio_dev_attr("current_timestamp_clock")

    @current_timestamp_clock.setter
    def current_timestamp_clock(self, value):
        self._set_iio_dev_attr_str("current_timestamp_clock", value)



# Copyright (C) 2019 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import re


def get_numbers(s):
    v = re.findall(r"[-+]?[.]?[\d]+(?:,\d\d\d)*[\.]?\d*(?:[eE][-+]?\d+)?", s)
    v = [float(i) for i in v]
    if len(v) == 1:
        v = v[0]
        if int(v) == v:
            v = int(v)
    return v


class attribute:
    def _get_iio_attr_str_multi_dev(self, channel_names, attr_name, output, ctrls):
        """ Get the same channel attribute across multiple devices
            which are assumed to be strings
        """
        if not isinstance(channel_names, list):
            channel_names = [channel_names]
        return {
            ctrl.name: [
                self._get_iio_attr_str(chan_name, attr_name, output, ctrl)
                for chan_name in channel_names
            ]
            for ctrl in ctrls
        }

    def _set_iio_attr_multi_dev(self, channel_names, attr_name, output, values, ctrls):
        """ Set the same channel attribute across multiple devices
            Unique parameters:
                values: type=list
                    Must be of length <= len(ctrls)*len(channel_names)
        """
        if len(values) > len(ctrls) * len(channel_names):
            raise Exception("Too many values to write")
        i = 0
        for ctrl in ctrls:
            for chan_name in channel_names:
                self._set_iio_attr(chan_name, attr_name, output, values[i], ctrl)
                i += 1

    def _set_iio_attr_float_multi_dev(
        self, channel_names, attr_name, output, values, ctrls
    ):
        """ Set the same float channel attribute(s) across multiple devices
            Unique parameters:
                values: type=list
                    Must be of length <= len(ctrls)*len(channel_names)
        """
        for i, value in enumerate(values):
            if isinstance(value, int):
                values[i] = float(value)
            if not isinstance(values[i], float):
                raise Exception("Values must be floats")
        self._set_iio_attr_multi_dev(channel_names, attr_name, output, values, ctrls)

    def _set_iio_attr(self, channel_name, attr_name, output, value, _ctrl=None):
        """ Set channel attribute """
        if _ctrl:
            channel = _ctrl.find_channel(channel_name, output)
        else:
            channel = self._ctrl.find_channel(channel_name, output)
        try:
            channel.attrs[attr_name].value = str(value)
        except Exception as ex:
            raise ex

    def _set_iio_attr_float(self, channel_name, attr_name, output, value, _ctrl=None):
        """ Set channel attribute with float """
        if isinstance(value, int):
            value = float(value)
        if not isinstance(value, float):
            raise Exception("Value must be a float")
        self._set_iio_attr(channel_name, attr_name, output, value, _ctrl)

    def _set_iio_attr_float_vec(
        self, channel_names, attr_name, output, values, _ctrl=None
    ):
        """ Set channel attribute with list of floats """
        if not isinstance(values, list):
            raise Exception("Value must be a list")
        for i, v in enumerate(values):
            self._set_iio_attr_float(channel_names[i], attr_name, output, v, _ctrl)

    def _set_iio_attr_int(self, channel_name, attr_name, output, value, _ctrl=None):
        """ Set channel attribute with int """
        if not isinstance(value, int):
            raise Exception("Value must be an int")
        self._set_iio_attr(channel_name, attr_name, output, value, _ctrl)

    def _set_iio_attr_int_vec(
        self, channel_names, attr_name, output, values, _ctrl=None
    ):
        """ Set channel attribute with list of ints """
        if not isinstance(values, list):
            raise Exception("Value must be a list")
        for i, v in enumerate(values):
            self._set_iio_attr_int(channel_names[i], attr_name, output, v, _ctrl)

    def _set_iio_attr_str_vec(
        self, channel_names, attr_name, output, values, _ctrl=None
    ):
        """ Set channel attribute with list of strings """
        if not isinstance(values, list):
            raise Exception("Value must be a list")
        for i, v in enumerate(list(values)):
            self._set_iio_attr(channel_names[i], attr_name, output, v, _ctrl)

    def _get_iio_attr_str(self, channel_name, attr_name, output, _ctrl=None):
        """ Get channel attribute as string """
        if _ctrl:
            channel = _ctrl.find_channel(channel_name, output)
        else:
            channel = self._ctrl.find_channel(channel_name, output)
        if not channel:
            raise Exception("No channel found with name: " + channel_name)
        return channel.attrs[attr_name].value

    def _get_iio_attr(self, channel_name, attr_name, output, _ctrl=None):
        """ Get channel attribute as number """
        return get_numbers(
            self._get_iio_attr_str(channel_name, attr_name, output, _ctrl)
        )

    def _get_iio_attr_vec(self, channel_names, attr_name, output, _ctrl=None):
        """ Get channel attributes as list of numbers """
        vals = []
        for chn in channel_names:
            v = self._get_iio_attr(chn, attr_name, output, _ctrl)
            vals.append(v)
        return vals

    def _get_iio_attr_str_vec(self, channel_names, attr_name, output, _ctrl=None):
        """ Get channel attributes as list of numbers """
        vals = []
        for chn in channel_names:
            v = self._get_iio_attr_str(chn, attr_name, output, _ctrl)
            vals.append(v)
        return vals

    def _set_iio_dev_attr_str(self, attr_name, value, _ctrl=None):
        """ Set device attribute with string """
        try:
            if _ctrl:
                _ctrl.attrs[attr_name].value = str(value)
            else:
                self._ctrl.attrs[attr_name].value = str(value)
        except Exception as ex:
            raise ex

    def _get_iio_dev_attr_str(self, attr_name, _ctrl=None):
        """ Get device attribute as string """
        if _ctrl:
            return _ctrl.attrs[attr_name].value
        else:
            return self._ctrl.attrs[attr_name].value

    def _set_iio_dev_attr(self, attr_name, value, _ctrl=None):
        """ Set device attribute """
        _dev = _ctrl or self._ctrl
        try:
            _dev.attrs[attr_name].value = str(value)
        except Exception as ex:
            raise ex

    def _get_iio_dev_attr(self, attr_name, _ctrl=None):
        """ Set device attribute as number """
        return get_numbers(self._get_iio_dev_attr_str(attr_name, _ctrl))

    def _set_iio_debug_attr_str(self, attr_name, value, _ctrl=None):
        """ Set debug attribute with string """
        try:
            if _ctrl:
                _ctrl.debug_attrs[attr_name].value = str(value)
            else:
                self._ctrl.debug_attrs[attr_name].value = str(value)
        except Exception as ex:
            raise ex

    def _get_iio_debug_attr_str(self, attr_name, _ctrl=None):
        """ Get debug attribute as string """
        if _ctrl:
            return _ctrl.debug_attrs[attr_name].value
        else:
            return self._ctrl.debug_attrs[attr_name].value

    def _get_iio_debug_attr(self, attr_name, _ctrl=None):
        """ Set debug attribute as number """
        return get_numbers(self._get_iio_debug_attr_str(attr_name, _ctrl))



# Copyright (C) 2021 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from adi.attribute import attribute
from adi.context_manager import context_manager


class admv8818(attribute, context_manager):
    """ADMV8818 2 GHz to 18 GHz, Digitally Tunable, High-Pass and Low-Pass Filter

    parameters:
        uri: type=string
            URI of IIO context with ADMV8818
    """

    _device_name = "admv8818"

    def __init__(self, uri="", device_name=""):

        context_manager.__init__(self, uri, self._device_name)

        self._device_name = device_name
        # Find the device
        self._ctrl = self._ctx.find_device(self._device_name)

        # Raise an exception if the device isn't found
        if not self._ctrl:
            raise Exception("ADMV8818 device not found")

    @property
    def low_pass_3db_frequency(self):
        """Get/Set the Low Pass 3dB Frequency in MHz"""
        return self._get_iio_attr(
            "altvoltage0", "filter_low_pass_3db_frequency", True, self._ctrl
        )

    @low_pass_3db_frequency.setter
    def low_pass_3db_frequency(self, value):
        """Get/Set the Low Pass 3dB Frequency in MHz"""
        self._set_iio_attr_int(
            "altvoltage0", "filter_low_pass_3db_frequency", True, int(value), self._ctrl
        )

    @property
    def high_pass_3db_frequency(self):
        """Get/Set the Low Pass 3dB Frequency in MHz"""
        return self._get_iio_attr(
            "altvoltage0", "filter_high_pass_3db_frequency", True, self._ctrl
        )

    @high_pass_3db_frequency.setter
    def high_pass_3db_frequency(self, value):
        """Get/Set the Low Pass 3dB Frequency in MHz"""
        self._set_iio_attr_int(
            "altvoltage0",
            "filter_high_pass_3db_frequency",
            True,
            int(value),
            self._ctrl,
        )

    @property
    def band_pass_bandwidth_3db_frequency(self):
        """Get/Set the Band Pass 3dB Frequency in MHz"""
        return self._get_iio_attr(
            "altvoltage0", "filter_band_pass_bandwidth_3db_frequency", True, self._ctrl
        )

    @band_pass_bandwidth_3db_frequency.setter
    def band_pass_bandwidth_3db_frequency(self, value):
        """Get/Set the Band Pass 3dB Frequency in MHz"""
        self._set_iio_attr_int(
            "altvoltage0",
            "filter_band_pass_bandwidth_3db_frequency",
            True,
            int(value),
            self._ctrl,
        )

    @property
    def band_pass_center_frequency(self):
        """Get/Set the Band Pass Center Frequency in MHz"""
        return self._get_iio_attr(
            "altvoltage0", "filter_band_pass_center_frequency", True, self._ctrl
        )

    @band_pass_center_frequency.setter
    def band_pass_center_frequency(self, value):
        """Get/Set the Band Pass Center Frequency in MHz"""
        self._set_iio_attr_int(
            "altvoltage0",
            "filter_band_pass_center_frequency",
            True,
            int(value),
            self._ctrl,
        )

    @property
    def mode_available(self):
        """Get available modes"""
        return self._get_iio_attr_str("altvoltage0", "mode_available", True, self._ctrl)

    @property
    def mode(self):
        """Get/Set mode"""
        return self._get_iio_attr_str("altvoltage0", "mode", True, self._ctrl)

    @mode.setter
    def mode(self, value):
        """Get/Set mode"""
        self._set_iio_attr("altvoltage0", "mode", True, value, self._ctrl)

    def reg_read(self, reg):
        """Direct Register Access via debugfs"""
        self._set_iio_debug_attr_str("direct_reg_access", reg, self._ctrl)
        return self._get_iio_debug_attr_str("direct_reg_access", self._ctrl)

    def reg_write(self, reg, value):
        """Direct Register Access via debugfs"""
        self._set_iio_debug_attr_str("direct_reg_access", f"{reg} {value}", self._ctrl)



# Copyright (C) 2019 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from adi.adrv9009 import adrv9009


class adrv9009_zu11eg(adrv9009):
    """ ADRV9009-ZU11EG System-On-Module

    parameters:
        uri: type=string
            URI of context with ADRV9009-ZU11EG
        jesd_monitor: type=boolean
            Boolean flag to enable JESD monitoring. jesd input is
            ignored otherwise.
        jesd: type=adi.jesd
            JESD object associated with ADRV9009-ZU11EG
    """

    _rx_channel_names = [
        "voltage0_i",
        "voltage0_q",
        "voltage1_i",
        "voltage1_q",
        "voltage2_i",
        "voltage2_q",
        "voltage3_i",
        "voltage3_q",
    ]
    _tx_channel_names = [
        "voltage0",
        "voltage1",
        "voltage2",
        "voltage3",
        "voltage4",
        "voltage5",
        "voltage6",
        "voltage7",
    ]
    _device_name = ""

    def __init__(self, uri="", jesd_monitor=False, jesd=None):

        adrv9009.__init__(self, uri=uri, jesd_monitor=jesd_monitor, jesd=jesd)
        self._ctrl_b = self._ctx.find_device("adrv9009-phy-b")
        self._clock_chip = self._ctx.find_device("hmc7044")
        self._clock_chip_carrier = self._ctx.find_device("hmc7044-car")
        # Used for multi-som sync
        self._clock_chip_ext = self._ctx.find_device("hmc7044-ext")

    def mcs_chips(self):
        """mcs_chips: MCS Synchronize both transceivers """
        try:
            _ = self.jesd204_fsm_ctrl
            # We're JESD204-fsm enabled - do nothing
        except:  # noqa: E722
            # Turn off continuous SYSREF, and enable GPI SYSREF request
            self._clock_chip.reg_write(0x5A, 0)
            for i in range(12):
                try:
                    self._set_iio_dev_attr_str("multichip_sync", i)
                except OSError:
                    pass
                try:
                    self._set_iio_dev_attr_str("multichip_sync", i, self._ctrl_b)
                except OSError:
                    pass

    @property
    def frequency_hopping_mode_chip_b(self):
        """frequency_hopping_mode_chip_b: Set Frequency Hopping Mode"""
        return self._get_iio_attr(
            "TRX_LO", "frequency_hopping_mode", True, self._ctrl_b
        )

    @frequency_hopping_mode_chip_b.setter
    def frequency_hopping_mode_chip_b(self, value):
        self._set_iio_attr(
            "TRX_LO", "frequency_hopping_mode", True, value, self._ctrl_b
        )

    @property
    def frequency_hopping_mode_en_chip_b(self):
        """frequency_hopping_mode_en: Enable Frequency Hopping Mode"""
        return self._get_iio_attr(
            "TRX_LO", "frequency_hopping_mode_enable", True, self._ctrl_b
        )

    @frequency_hopping_mode_en_chip_b.setter
    def frequency_hopping_mode_en_chip_b(self, value):
        self._set_iio_attr(
            "TRX_LO", "frequency_hopping_mode_enable", True, value, self._ctrl_b
        )

    @property
    def calibrate_rx_phase_correction_en_chip_b(self):
        """calibrate_rx_phase_correction_en: Enable RX Phase Correction Calibration"""
        return self._get_iio_dev_attr("calibrate_rx_phase_correction_en", self._ctrl_b)

    @calibrate_rx_phase_correction_en_chip_b.setter
    def calibrate_rx_phase_correction_en_chip_b(self, value):
        self._set_iio_dev_attr_str(
            "calibrate_rx_phase_correction_en", value, self._ctrl_b
        )

    @property
    def calibrate_rx_qec_en_chip_b(self):
        """calibrate_rx_qec_en_chip_b: Enable RX QEC Calibration"""
        return self._get_iio_dev_attr("calibrate_rx_qec_en", self._ctrl_b)

    @calibrate_rx_qec_en_chip_b.setter
    def calibrate_rx_qec_en_chip_b(self, value):
        self._set_iio_dev_attr_str("calibrate_rx_qec_en", value, self._ctrl_b)

    @property
    def calibrate_tx_qec_en_chip_b(self):
        """calibrate_tx_qec_en_chip_b: Enable TX QEC Calibration"""
        return self._get_iio_dev_attr("calibrate_tx_qec_en", self._ctrl_b)

    @calibrate_tx_qec_en_chip_b.setter
    def calibrate_tx_qec_en_chip_b(self, value):
        self._set_iio_dev_attr_str("calibrate_tx_qec_en", value, self._ctrl_b)

    @property
    def calibrate_chip_b(self):
        """calibrate_chip_b: Trigger Calibration"""
        return self._get_iio_dev_attr("calibrate", self._ctrl_b)

    @calibrate_chip_b.setter
    def calibrate_chip_b(self, value):
        self._set_iio_dev_attr_str("calibrate", value, self._ctrl_b)

    @property
    def gain_control_mode_chan0_chip_b(self):
        """gain_control_mode_chan0_chip_b: Mode of receive path AGC. Options are:
        slow_attack, manual"""
        return self._get_iio_attr_str(
            "voltage0", "gain_control_mode", False, self._ctrl_b
        )

    @gain_control_mode_chan0_chip_b.setter
    def gain_control_mode_chan0_chip_b(self, value):
        self._set_iio_attr("voltage0", "gain_control_mode", False, value, self._ctrl_b)

    @property
    def gain_control_mode_chan1_chip_b(self):
        """gain_control_mode_chan1_chip_b: Mode of receive path AGC. Options are:
        slow_attack, manual"""
        return self._get_iio_attr_str(
            "voltage1", "gain_control_mode", False, self._ctrl_b
        )

    @gain_control_mode_chan1_chip_b.setter
    def gain_control_mode_chan1_chip_b(self, value):
        self._set_iio_attr("voltage1", "gain_control_mode", False, value, self._ctrl_b)

    @property
    def rx_hardwaregain_chan0_chip_b(self):
        """rx_hardwaregain: Gain applied to RX path channel 0. Only applicable when
        gain_control_mode is set to 'manual'"""
        return self._get_iio_attr("voltage0", "hardwaregain", False, self._ctrl_b)

    @rx_hardwaregain_chan0_chip_b.setter
    def rx_hardwaregain_chan0_chip_b(self, value):
        if self.gain_control_mode_chan0_chip_b == "manual":
            self._set_iio_attr("voltage0", "hardwaregain", False, value, self._ctrl_b)

    @property
    def rx_hardwaregain_chan1_chip_b(self):
        """rx_hardwaregain: Gain applied to RX path channel 1. Only applicable when
        gain_control_mode is set to 'manual'"""
        return self._get_iio_attr("voltage1", "hardwaregain", False, self._ctrl_b)

    @rx_hardwaregain_chan1_chip_b.setter
    def rx_hardwaregain_chan1_chip_b(self, value):
        if self.gain_control_mode_chan1_chip_b == "manual":
            self._set_iio_attr("voltage1", "hardwaregain", False, value, self._ctrl_b)

    @property
    def tx_hardwaregain_chan0_chip_b(self):
        """tx_hardwaregain: Attenuation applied to TX path channel 0"""
        return self._get_iio_attr("voltage0", "hardwaregain", True, self._ctrl_b)

    @tx_hardwaregain_chan0_chip_b.setter
    def tx_hardwaregain_chan0_chip_b(self, value):
        self._set_iio_attr("voltage0", "hardwaregain", True, value, self._ctrl_b)

    @property
    def tx_hardwaregain_chan1_chip_b(self):
        """tx_hardwaregain: Attenuation applied to TX path channel 1"""
        return self._get_iio_attr("voltage1", "hardwaregain", True, self._ctrl_b)

    @tx_hardwaregain_chan1_chip_b.setter
    def tx_hardwaregain_chan1_chip_b(self, value):
        self._set_iio_attr("voltage1", "hardwaregain", True, value, self._ctrl_b)

    @property
    def rx_rf_bandwidth_chip_b(self):
        """rx_rf_bandwidth: Bandwidth of front-end analog filter of RX path"""
        return self._get_iio_attr("voltage0", "rf_bandwidth", False, self._ctrl_b)

    @property
    def tx_rf_bandwidth_chip_b(self):
        """tx_rf_bandwidth: Bandwidth of front-end analog filter of TX path"""
        return self._get_iio_attr("voltage0", "rf_bandwidth", True, self._ctrl_b)

    @property
    def rx_sample_rate_chip_b(self):
        """rx_sample_rate: Sample rate RX path in samples per second"""
        return self._get_iio_attr("voltage0", "sampling_frequency", False, self._ctrl_b)

    @property
    def tx_sample_rate_chip_b(self):
        """tx_sample_rate: Sample rate TX path in samples per second"""
        return self._get_iio_attr("voltage0", "sampling_frequency", True, self._ctrl_b)

    @property
    def trx_lo_chip_b(self):
        """trx_lo: Carrier frequency of TX and RX path"""
        return self._get_iio_attr("altvoltage0", "frequency", True, self._ctrl_b)

    @trx_lo_chip_b.setter
    def trx_lo_chip_b(self, value):
        self._set_iio_attr("altvoltage0", "frequency", True, value, self._ctrl_b)



# Copyright (C) 2021 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from adi.attribute import attribute
from adi.context_manager import context_manager


class adf4159(context_manager, attribute):

    """ ADF4159 is a 13 GHz Fractional-N Frequency Synthesizer """

    _device_name = "adf4159"

    def __init__(self, uri=""):

        context_manager.__init__(self, uri, self._device_name)
        # Find the device
        self._ctrl = self._ctx.find_device(self._device_name)
        # Raise an exception if the device isn't found
        if not self._ctrl:
            raise Exception("ADF4159 device not found")

    @property
    def ramp_mode(self):
        """Get/Set the Ramp output mode."""
        return self._get_iio_attr_str("altvoltage0", "ramp_mode", True, self._ctrl)

    @ramp_mode.setter
    def ramp_mode(self, value):
        """Get/Set the Ramp output mode."""

        valid = self._get_iio_attr_str(
            "altvoltage0", "ramp_mode_available", True, self._ctrl
        )
        if value not in valid:

            raise ValueError(
                f'ramp_mode of "{value}" is invalid. Valid options: "{valid}"'
            )

        self._set_iio_attr("altvoltage0", "ramp_mode", True, value, self._ctrl)

    @property
    def enable(self):
        """Get/Set the enable status of the RF output."""
        return bool(
            int(self._get_iio_attr("altvoltage0", "powerdown", True, self._ctrl))
        )

    @enable.setter
    def enable(self, value):
        """Get/Set the enable status of the RF output."""
        self._set_iio_attr("altvoltage0", "powerdown", True, int(value), self._ctrl)

    @property
    def frequency(self):
        """Get/Set the Output Frequency of PLL."""
        return self._get_iio_attr("altvoltage0", "frequency", True, self._ctrl)

    @frequency.setter
    def frequency(self, value):
        """Get/Set the Output Frequency of PLL."""
        self._set_iio_attr("altvoltage0", "frequency", True, value, self._ctrl)

    @property
    def freq_dev_range(self):
        """Get/Set the PLL frequency deviation range."""
        return self._get_iio_attr(
            "altvoltage0", "frequency_deviation_range", True, self._ctrl
        )

    @freq_dev_range.setter
    def freq_dev_range(self, value):
        """Get/Set the PLL frequency deviation range."""
        self._set_iio_attr(
            "altvoltage0", "frequency_deviation_range", True, value, self._ctrl
        )

    @property
    def freq_dev_step(self):
        """Get/Set the PLL frequency deviation step."""
        return self._get_iio_attr(
            "altvoltage0", "frequency_deviation_step", True, self._ctrl
        )

    @freq_dev_step.setter
    def freq_dev_step(self, value):
        """Get/Set the PLL frequency deviation step."""
        self._set_iio_attr(
            "altvoltage0", "frequency_deviation_step", True, (1 + value), self._ctrl
        )

    @property
    def freq_dev_time(self):
        """Get/Set the PLL frequency deviation time."""
        return self._get_iio_attr(
            "altvoltage0", "frequency_deviation_time", True, self._ctrl
        )

    @freq_dev_time.setter
    def freq_dev_time(self, value):
        """Get/Set the PLL frequency deviation time."""
        self._set_iio_attr(
            "altvoltage0", "frequency_deviation_time", True, value, self._ctrl
        )



# Copyright (C) 2021 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from adi.attribute import attribute
from adi.context_manager import context_manager
from adi.rx_tx import rx


class adpd1080(rx, context_manager):
    """ADPD1080 photo-electronic device."""

    channel = []  # type: ignore
    _device_name = ""

    def __init__(self, uri="", device_index=0):
        """ADPD1080 class constructor."""
        context_manager.__init__(self, uri, self._device_name)

        compatible_parts = ["adpd1080"]

        self._ctrl = None
        index = 0

        # Selecting the device_index-th device from the adpd188 family as working device.
        for device in self._ctx.devices:
            if device.name in compatible_parts:
                if index == device_index:
                    self._ctrl = device
                    self._rxadc = device
                    break
                else:
                    index += 1

        # dynamically get channels and sorting them after the color
        # self._ctrl.channels.sort(key=lambda x: str(x.id[14:]))

        for ch in self._ctrl._channels:
            name = ch._id
            self._rx_channel_names.append(name)
            self.channel.append(self._channel(self._ctrl, name))
        rx.__init__(self)
        self.rx_buffer_size = 16

    def rx(self):
        buff = super().rx()
        del self._rx__rxbuf
        return buff

    @property
    def sample_rate(self):
        """Sets sampling frequency of the ADPD1080."""
        return self._get_iio_dev_attr_str("sampling_frequency", self._rxadc)

    @sample_rate.setter
    def sample_rate(self, value):
        self._set_iio_dev_attr_str("sampling_frequency", value, self._rxadc)

    class _channel(attribute):
        """ADPD1080 channel."""

        def __init__(self, ctrl, channel_name):
            self.name = channel_name
            self._ctrl = ctrl

        @property
        def raw(self):
            """ADPD1080 channel raw value."""
            return self._get_iio_attr(self.name, "raw", False)

        @property
        def offset(self):
            """ADPD1080 channel offset."""
            return self._get_iio_attr(self.name, "offset", False)

        @offset.setter
        def offset(self, value):
            self._set_iio_attr(self.name, "offset", False, value)



# Copyright (C) 2020 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from math import pi, sin

from adi.attribute import attribute
from adi.context_manager import context_manager


class adar1000(attribute, context_manager):
    """ ADAR1000 Beamformer

    parameters:
        uri: type=string
            Optional parameter for the URI of IIO context with ADAR1000(s). If
            not given,
        context: type=iio.Context
            Optional parameter for IIO Context handle for the device. Don't use if
            instantiating the adar1000 class directly. The adar1000_array class
            will handle this if creating an array instance.
        chip_id: type=string
            Required string identifying the desired chip select and hardware ID
            for the ADAR1000. If creating a single adar1000 instance, you can
            typically leave the default value of "csb1_chip1" as long as the
            device tree label matches. If creating an adar1000_array instance,
            the array class will handle the instantiation of individual adar1000
            handles.
        device_number: type=int
            Required integer indicating the device number in the array. If creating
            a single adar1000 instance, this value should be 1. If creating an
            adar1000_array instance, the array class will handle the instantiation
            of individual adar1000 handles.
        array_element_map: type=list[list[int]]
            Required list of lists with the map of where the array elements are
            located in the physical array. Each entry in the map represents a
            row of elements referenced by element number. For example, a map:

                | [[1, 5, 9, 13],
                | [2, 6, 10, 14],
                | [3, 7, 11, 15],
                | [4, 8, 12, 16]]

            represents an array of 16 elements (4 ADAR1000s) in a 4x4 array.
            If creating a single adar1000 instance, the elements should be 1-4 in
            whatever configuration the physical array is (1x4, 2x2, 4x1, etc.). If
            creating an adar1000_array instance, the array class will handle the
            instantiation of individual adar1000 handles.
        channel_element_map: type=list[int]
            Required list of integers relating the array element numbers to the
            channels of the ADAR1000 instance. Each number in the list is the
            element number in the larger array, in order of the ADAR1000's channels.
            For example, a list [10, 14, 13, 9] indicates that the ADAR1000's
            channels are the following elements in the full array:

                | Channel 1: Element # 10
                | Channel 2: Element # 14
                | Channel 3: Element # 13
                | Channel 4: Element # 9

            If creating a single adar1000 instance, the elements should be 1-4 in
            order of the ADAR1000's channels related to the array elements. If
            creating an adar1000_array instance, the array class will handle the
            instantiation of individual adar1000 handles.
    """

    _device_name = ""
    _BIAS_CODE_TO_VOLTAGE_SCALE = -0.018824

    class adar1000_channel:
        """Class for each channel of the ADAR1000. This class is not meant
        to be instantiated directly. adar1000 objects will create their
        own handles of this class, one for each channel

        parameters:
            adar1000_parent: type=adar1000
                Parent ADAR1000 instance
            adar_channel: type=int
                Channel number of the parent corresponding to this channel
            array_element: type=int
                Element number of the array corresponding to this channel
            row: int
                Row number in the array
            column: int
                Column number in the array
        """

        def __init__(self, adar_parent, adar_channel, array_element, row, column):
            self._adar1000_parent = adar_parent
            self._adar1000_channel = adar_channel
            self._array_element_number = array_element
            self._row = row
            self._column = column

            self._BIAS_CODE_TO_VOLTAGE_SCALE = adar1000._BIAS_CODE_TO_VOLTAGE_SCALE

        def __repr__(self):
            """ Representation of the ADAR1000 element class """
            return f"ADAR1000 array element #{self.array_element_number}"

        """ Channel Properties """

        @property
        def adar1000_channel(self):
            return self._adar1000_channel

        @property
        def adar1000_parent(self):
            return self._adar1000_parent

        @property
        def array_element_number(self):
            """ Element number in the array """
            return self._array_element_number

        @property
        def column(self):
            """ Element column number in the array """
            return self._column

        @property
        def _detector_enable(self):
            """ Get/Set the detector enable bit for the associated channel """
            return bool(
                self.adar1000_parent._get_iio_attr(
                    f"voltage{self.adar1000_channel}", "detector_en", True
                )
            )

        @_detector_enable.setter
        def _detector_enable(self, value):
            """ Get/Set Channel 1 detector enable bit """
            self.adar1000_parent._set_iio_attr(
                f"voltage{self.adar1000_channel}", "detector_en", True, int(value)
            )

        @property
        def detector_power(self):
            """ Get the detector power reading for the associated channel """

            self._detector_enable = True

            readback = self.adar1000_parent._get_iio_attr(
                f"voltage{self.adar1000_channel}", "raw", True
            )

            self._detector_enable = False

            return readback

        @property
        def pa_bias_off(self):
            """ Get/Set PA_BIAS_OFF in voltage for the associated channel """
            dac_code = self.adar1000_parent._get_iio_attr(
                f"voltage{self.adar1000_channel}", "pa_bias_off", True
            )
            return dac_code * self._BIAS_CODE_TO_VOLTAGE_SCALE

        @pa_bias_off.setter
        def pa_bias_off(self, value):
            """ Get/Set PA_BIAS_OFF in voltage for the associated channel """
            dac_code = int(value / self._BIAS_CODE_TO_VOLTAGE_SCALE)
            self.adar1000_parent._set_iio_attr(
                f"voltage{self.adar1000_channel}", "pa_bias_off", True, dac_code
            )

        @property
        def pa_bias_on(self):
            """ Get/Set PA_BIAS_ON in voltage for the associated channel """
            dac_code = self.adar1000_parent._get_iio_attr(
                f"voltage{self.adar1000_channel}", "pa_bias_on", True
            )
            return dac_code * self._BIAS_CODE_TO_VOLTAGE_SCALE

        @pa_bias_on.setter
        def pa_bias_on(self, value):
            """ Get/Set PA_BIAS_ON in voltage for the associated channel """
            dac_code = int(value / self._BIAS_CODE_TO_VOLTAGE_SCALE)
            self.adar1000_parent._set_iio_attr(
                f"voltage{self.adar1000_channel}", "pa_bias_on", True, dac_code
            )

        @property
        def row(self):
            """ Element row number in the array """
            return self._row

        @property
        def rx_attenuator(self):
            """ Get/Set Rx Attenuator state for the associated channel """
            return bool(
                1
                - self.adar1000_parent._get_iio_attr(
                    f"voltage{self.adar1000_channel}", "attenuation", False
                )
            )

        @rx_attenuator.setter
        def rx_attenuator(self, value):
            """ Get/Set Rx Attenuator state for the associated channel """
            self.adar1000_parent._set_iio_attr(
                f"voltage{self.adar1000_channel}", "attenuation", False, 1 - int(value)
            )

        @property
        def rx_beam_state(self):
            """Get/Set the Channel Rx beam position used by RAM when
            all channels point to individual states. Valid states are 0-120."""
            return self.adar1000_parent._get_iio_attr(
                f"voltage{self.adar1000_channel}", "beam_pos_load", False
            )

        @rx_beam_state.setter
        def rx_beam_state(self, value):
            """Get/Set the Channel Rx beam position used by RAM when
            all channels point to individual states. Valid states are 0-120."""
            self.adar1000_parent._set_iio_attr(
                f"voltage{self.adar1000_channel}", "beam_pos_load", False, value
            )

        @property
        def rx_enable(self):
            """ Get/Set the Rx enable state for the associated channel """
            return bool(
                1
                - self.adar1000_parent._get_iio_attr(
                    f"voltage{self.adar1000_channel}", "powerdown", False
                )
            )

        @rx_enable.setter
        def rx_enable(self, value):
            """ Get/Set the Rx enable state for the associated channel """
            self.adar1000_parent._set_iio_attr(
                f"voltage{self.adar1000_channel}", "powerdown", False, 1 - int(value)
            )

        @property
        def rx_gain(self):
            """ Get/Set the Rx Gain for the associated channel """
            return self.adar1000_parent._get_iio_attr(
                f"voltage{self.adar1000_channel}", "hardwaregain", False
            )

        @rx_gain.setter
        def rx_gain(self, value):
            """ Get/Set the Rx Gain for the associated channel """
            self.adar1000_parent._set_iio_attr(
                f"voltage{self.adar1000_channel}", "hardwaregain", False, value
            )

        @property
        def rx_phase(self):
            """ Get/Set the Rx Phase for the associated channel """
            return self.adar1000_parent._get_iio_attr(
                f"voltage{self.adar1000_channel}", "phase", False
            )

        @rx_phase.setter
        def rx_phase(self, value):
            """ Get/Set the Rx Phase for the associated channel """
            self.adar1000_parent._set_iio_attr(
                f"voltage{self.adar1000_channel}", "phase", False, value
            )

        @property
        def tx_attenuator(self):
            """ Get/Set the Tx Attenuator state for the associated channel """
            return bool(
                1
                - self.adar1000_parent._get_iio_attr(
                    f"voltage{self.adar1000_channel}", "attenuation", True
                )
            )

        @tx_attenuator.setter
        def tx_attenuator(self, value):
            """ Get/Set the Tx Attenuator state for the associated channel """
            self.adar1000_parent._set_iio_attr(
                f"voltage{self.adar1000_channel}", "attenuation", True, 1 - int(value)
            )

        @property
        def tx_beam_state(self):
            """Get/Set the Channel Tx beam position used by RAM when
            all channels point to individual states. Valid states are 0-120."""
            return self.adar1000_parent._get_iio_attr(
                f"voltage{self.adar1000_channel}", "beam_pos_load", True
            )

        @tx_beam_state.setter
        def tx_beam_state(self, value):
            """Get/Set the Channel Tx beam position used by RAM when
            all channels point to individual states. Valid states are 0-120."""
            self.adar1000_parent._set_iio_attr(
                f"voltage{self.adar1000_channel}", "beam_pos_load", True, value
            )

        @property
        def tx_enable(self):
            """ Get/Set the Tx enable state for the associated channel """
            return bool(
                1
                - self.adar1000_parent._get_iio_attr(
                    f"voltage{self.adar1000_channel}", "powerdown", True
                )
            )

        @tx_enable.setter
        def tx_enable(self, value):
            """ Get/Set the Tx enable state for the associated channel """
            self.adar1000_parent._set_iio_attr(
                f"voltage{self.adar1000_channel}", "powerdown", True, 1 - int(value)
            )

        @property
        def tx_gain(self):
            """ Get/Set the Tx Gain for the associated channel """
            return self.adar1000_parent._get_iio_attr(
                f"voltage{self.adar1000_channel}", "hardwaregain", True
            )

        @tx_gain.setter
        def tx_gain(self, value):
            """ Get/Set the Tx Gain for the associated channel """
            self.adar1000_parent._set_iio_attr(
                f"voltage{self.adar1000_channel}", "hardwaregain", True, value
            )

        @property
        def tx_phase(self):
            """ Get/Set the Tx Phase for the associated channel """
            return self.adar1000_parent._get_iio_attr(
                f"voltage{self.adar1000_channel}", "phase", True
            )

        @tx_phase.setter
        def tx_phase(self, value):
            """ Get/Set the Tx Phase for the associated channel """
            self.adar1000_parent._set_iio_attr(
                f"voltage{self.adar1000_channel}", "phase", True, value
            )

        """ Public Methods """

        def save_rx_beam(self, state, attenuator, gain, phase):
            """Save a beam to an Rx memory position

            parameters:
                state: int
                    State number to save. Valid options are 0 to 120
                attenuator: bool
                    Attenuator state for the beam position. True means the attenuator is in place.
                gain: int
                    Gain value for the beam position. Valid settings are 0 to 127.
                phase: float
                    Phase value for the beam position.
            """
            save_string = f"{state}, {1 - int(attenuator)}, {gain}, {phase}"

            self.adar1000_parent._set_iio_attr(
                f"voltage{self.adar1000_channel}", "beam_pos_save", False, save_string
            )

        def save_tx_beam(self, state, attenuator, gain, phase):
            """Save a beam to a Tx memory position

            parameters:
                state: int
                    State number to save. Valid options are 0 to 120
                attenuator: bool
                    Attenuator state for the beam position. True means the attenuator is in place.
                gain: int
                    Gain value for the beam position. Valid settings are 0 to 127.
                phase: float
                    Phase value for the beam position.
            """
            save_string = f"{state}, {1 - int(attenuator)}, {gain}, {phase}"

            self.adar1000_parent._set_iio_attr(
                f"voltage{self.adar1000_channel}", "beam_pos_save", True, save_string
            )

    def __init__(
        self,
        uri="",
        context=None,
        chip_id="csb1_chip1",
        device_number=1,
        array_element_map=None,
        channel_element_map=None,
    ):

        # Use the given context if possible, otherwise lookup the context
        if context is None:
            context_manager.__init__(self, uri, self._device_name)
        else:
            self._ctx = context

        self._ctrl = None
        self._chip_id = chip_id
        self._array_device_number = device_number

        # Check for the presence of array and channel element maps
        if array_element_map is None:
            raise Exception('"array_element_map" argument must be provided!')
        if channel_element_map is None:
            raise Exception('"channel_element_map" argument must be provided!')

        # Look for the matching device in the context
        for dev in self._ctx.devices:
            if (
                "label" in dev.attrs
                and dev.attrs["label"].value.lower() == chip_id.lower()
            ) or (
                hasattr(dev, "label")
                and dev.label
                and dev.label.lower() == chip_id.lower()
            ):
                self._ctrl = dev
                break

        # Raise an exception if the device isn't found
        if not self._ctrl:
            raise Exception(f"Device not found for {chip_id}")

        # Determine the row and column numbers for the elements using the maps given
        element_numbers = channel_element_map
        element_rows_cols = {}
        for element_number in element_numbers:
            for r, row in enumerate(array_element_map):
                for c, map_value in enumerate(row):
                    if map_value == element_number:
                        element_rows_cols[element_number] = (r, c)

        # Create the channel instances
        self._channels = []
        for i in range(4):
            element_number = element_numbers[i]
            row, column = element_rows_cols[element_number]
            self._channels.append(
                self.adar1000_channel(self, i, element_number, row, column)
            )

    def __repr__(self):
        """ Representation of the ADAR1000 class """
        return (
            f"ADAR1000 #{self.array_device_number} controlling array elements:\n"
            f"\t{', '.join([str(channel.array_element_number) for channel in self.channels])}"
        )

    """ Device Properties """

    @property
    def array_device_number(self):
        """ Device number in the array """
        return self._array_device_number

    @property
    def beam_mem_enable(self):
        """ Get/Set enable bit for RAM control vs. SPI control of the beam state """
        return bool(self._get_iio_dev_attr("beam_mem_enable", self._ctrl))

    @beam_mem_enable.setter
    def beam_mem_enable(self, value):
        """ Get/Set enable bit for RAM control vs. SPI control of the beam state """
        self._set_iio_dev_attr_str("beam_mem_enable", int(value), self._ctrl)

    @property
    def bias_dac_enable(self):
        """ Get/Set enable for bias DACs """
        return bool(self._get_iio_dev_attr("bias_enable", self._ctrl))

    @bias_dac_enable.setter
    def bias_dac_enable(self, value):
        """ Get/Set enable for bias DACs """
        self._set_iio_dev_attr_str("bias_enable", int(value), self._ctrl)

    @property
    def bias_dac_mode(self):
        """
        Get/Set BIAS_CTRL bit (Register 0x30[6]) which controls whether the bias DACs stay at "ON" value,
        or toggle with respect to T/R state.
        """
        value = bool(self._get_iio_dev_attr("bias_ctrl", self._ctrl))
        if value:
            return "toggle"
        else:
            return "on"

    @bias_dac_mode.setter
    def bias_dac_mode(self, value):
        """
        Get/Set BIAS_CTRL bit (Register 0x30[6]) which controls whether the bias DACs stay at "ON" value,
        or toggle with respect to T/R state.
        """
        if value.lower() == "toggle":
            set_value = 1
        elif value.lower() == "on":
            set_value = 0
        else:
            raise ValueError('bias_dac_mode should be "toggle" or "on"')

        self._set_iio_dev_attr_str("bias_ctrl", set_value, self._ctrl)

    @property
    def bias_mem_enable(self):
        """ Get/Set enable bit for RAM control vs. SPI control of the bias state """
        return bool(self._get_iio_dev_attr("bias_mem_enable", self._ctrl))

    @bias_mem_enable.setter
    def bias_mem_enable(self, value):
        """ Get/Set enable bit for RAM control vs. SPI control of the bias state """
        self._set_iio_dev_attr_str("bias_mem_enable", int(value), self._ctrl)

    @property
    def channel1(self):
        """ Handle for the ADAR1000's channel 1 object """
        return self._channels[0]

    @property
    def channel2(self):
        """ Handle for the ADAR1000's channel 2 object """
        return self._channels[1]

    @property
    def channel3(self):
        """ Handle for the ADAR1000's channel 3 object """
        return self._channels[2]

    @property
    def channel4(self):
        """ Handle for the ADAR1000's channel 4 object """
        return self._channels[3]

    @property
    def channels(self):
        """ List of the ADAR1000's channels, in order from 1 to 4 """
        return self._channels

    @property
    def chip_id(self):
        """
        Chip ID including CSB and hardware address. Of the form "csbX_chipX" where csbX indicates the CSB
        line for the IC and chipX indicating the hardware address of the chip, 1-4.
        """
        return self._chip_id

    @property
    def common_mem_enable(self):
        """
        Get/Set the CHX_RAM_BYPASS bits to use either a common beam state for all channels set by registers 0x39
        and 0x3A, or individual beam states set by registers 0x3D to 0x44.
        """
        return bool(self._get_iio_dev_attr("common_mem_enable", self._ctrl))

    @common_mem_enable.setter
    def common_mem_enable(self, value):
        """
        Get/Set the CHX_RAM_BYPASS bits to use either a common beam state for all channels set by registers 0x39
        and 0x3A, or individual beam states set by registers 0x3D to 0x44.
        """
        self._set_iio_dev_attr_str("bias_enable", int(value), self._ctrl)

    @property
    def common_rx_beam_state(self):
        """Get/Set the Rx beam position used by RAM when all
        channels point to a common state. Valid states are 0-120."""
        return self._get_iio_dev_attr_str("static_rx_beam_pos_load", self._ctrl)

    @common_rx_beam_state.setter
    def common_rx_beam_state(self, value):
        """Get/Set the Rx beam position used by RAM when all
        channels point to a common state. Valid states are 0-120."""
        self._set_iio_dev_attr_str("static_rx_beam_pos_load", value, self._ctrl)

    @property
    def common_tx_beam_state(self):
        """Get/Set the Tx beam position used by RAM when all
        channels point to a common state. Valid states are 0-120."""
        return self._get_iio_dev_attr_str("static_tx_beam_pos_load", self._ctrl)

    @common_tx_beam_state.setter
    def common_tx_beam_state(self, value):
        """Get/Set the Tx beam position used by RAM when all
        channels point to a common state. Valid states are 0-120."""
        self._set_iio_dev_attr_str("static_tx_beam_pos_load", value, self._ctrl)

    @property
    def external_tr_pin(self):
        """ Get/Set which external T/R switch driver is used ("positive" = TR_SW_POS, "negative" = TR_SW_NEG) """
        value = bool(self._get_iio_dev_attr("sw_drv_tr_mode_sel", self._ctrl))
        if value:
            return "negative"
        else:
            return "positive"

    @external_tr_pin.setter
    def external_tr_pin(self, value):
        """ Get/Set which external T/R switch driver is used ("positive" = TR_SW_POS, "negative" = TR_SW_NEG) """
        if value.lower() == "negative":
            set_value = 1
        elif value.lower() == "positive":
            set_value = 0
        else:
            raise ValueError('external_tr_pin should be "positive" or "negative"')

        self._set_iio_dev_attr_str("sw_drv_tr_mode_sel", set_value, self._ctrl)

    @property
    def external_tr_polarity(self):
        """
        Get/Set polarity of the T/R switch driver compared to the T/R state of the ADAR1000.
        True outputs 0V in Rx mode, False outputs either 3.3V or -5V, depending on which T/R switch driver is enabled.
        """
        return bool(self._get_iio_dev_attr("sw_drv_tr_state", self._ctrl))

    @external_tr_polarity.setter
    def external_tr_polarity(self, value):
        """
        Get/Set polarity of the T/R switch driver compared to the T/R state of the ADAR1000.
        True outputs 0V in Rx mode, False outputs either 3.3V or -5V, depending on which T/R switch driver is enabled.
        """
        self._set_iio_dev_attr_str("sw_drv_tr_state", int(value), self._ctrl)

    @property
    def lna_bias_off(self):
        """ Get/Set LNA_BIAS_OFF in voltage """
        dac_code = self._get_iio_dev_attr("lna_bias_off", self._ctrl)
        return dac_code * self._BIAS_CODE_TO_VOLTAGE_SCALE

    @lna_bias_off.setter
    def lna_bias_off(self, value):
        """ Get/Set LNA_BIAS_OFF in voltage """
        dac_code = int(value / self._BIAS_CODE_TO_VOLTAGE_SCALE)
        self._set_iio_dev_attr_str("lna_bias_off", dac_code, self._ctrl)

    @property
    def lna_bias_on(self):
        """ Get/Set LNA_BIAS_ON in voltage """
        dac_code = self._get_iio_dev_attr("lna_bias_on", self._ctrl)
        return dac_code * self._BIAS_CODE_TO_VOLTAGE_SCALE

    @lna_bias_on.setter
    def lna_bias_on(self, value):
        """ Get/Set LNA_BIAS_ON in voltage """
        dac_code = int(value / self._BIAS_CODE_TO_VOLTAGE_SCALE)
        self._set_iio_dev_attr_str("lna_bias_on", dac_code, self._ctrl)

    @property
    def lna_bias_out_enable(self):
        """ Get/Set enable for LNA bias DAC output. Disable to allow for always-on self-biased LNAs """
        return bool(self._get_iio_dev_attr("lna_bias_out_enable", self._ctrl))

    @lna_bias_out_enable.setter
    def lna_bias_out_enable(self, value):
        """ Get/Set enable for LNA bias DAC output. Disable to allow for always-on self-biased LNAs """
        self._set_iio_dev_attr_str("lna_bias_out_enable", int(value), self._ctrl)

    @property
    def mode(self):
        """ Get/Set the mode of operation for the device. Valid options are "rx", "tx", and "disabled" """

        # There are different paths for spi vs. external T/R control

        if self.tr_source == "external":
            return self._tr

        else:
            # Check for Rx mode
            if self._rx_enable and not self._tx_enable:
                if self._tr == "rx":
                    return "rx"
                else:
                    return "disabled"

            # Check for Tx mode
            elif self._tx_enable and not self._rx_enable:
                if self._tr == "tx":
                    return "tx"
                else:
                    return "disabled"

            # Check mode if both device enables are high
            elif self._rx_enable and self._tx_enable:
                return self._tr

            # Device is disabled
            else:
                return "disabled"

    @mode.setter
    def mode(self, value):
        """ Get/Set the mode of operation for the device. Valid options are "rx", "tx", and "disabled" """
        mode = value.strip().lower()
        if mode not in ("rx", "tx", "disabled"):
            raise ValueError(
                'Mode of operation must be either "rx", "tx", or "disabled"'
            )

        if mode == "disabled":
            self.tr_source = "spi"
            self._rx_enable = False
            self._tx_enable = False

        else:
            # Set the T/R state to the correct mode
            self._tr = mode

            if mode == "rx":
                self._rx_enable = True
                self._tx_enable = False

            # Tx mode
            else:
                self._rx_enable = False
                if self.tr_source == "spi":
                    self._tx_enable = True

    @property
    def pol_state(self):
        """ Get/Set polarity switch state. True outputs -5V, False outputs 0V """
        return bool(self._get_iio_dev_attr("pol", self._ctrl))

    @pol_state.setter
    def pol_state(self, value):
        """ Get/Set polarity switch state. True outputs -5V, False outputs 0V """
        self._set_iio_dev_attr_str("pol", int(value), self._ctrl)

    @property
    def pol_switch_enable(self):
        """ Get/Set polarity switch driver enable state """
        return bool(self._get_iio_dev_attr("sw_drv_en_pol", self._ctrl))

    @pol_switch_enable.setter
    def pol_switch_enable(self, value):
        """ Get/Set polarity switch driver enable state """
        self._set_iio_dev_attr_str("sw_drv_en_pol", int(value), self._ctrl)

    @property
    def rx_bias_state(self):
        """ Get/Set the Rx bias memory position when loading from RAM. Valid states are 1-7. """
        return self._get_iio_attr("voltage0", "bias_set_load", False, self._ctrl)

    @rx_bias_state.setter
    def rx_bias_state(self, value):
        """ Get/Set the Rx bias memory position when loading from RAM. Valid states are 1-7. """
        self._set_iio_attr("voltage0", "bias_set_load", False, value, self._ctrl)

    @property
    def _rx_enable(self):
        """ Get/Set enable for entire Rx chain. Should not be used directly, the mode property controls this. """
        return bool(self._get_iio_dev_attr("rx_en", self._ctrl))

    @_rx_enable.setter
    def _rx_enable(self, value):
        """ Get/Set enable for entire Rx chain. Should not be used directly, the mode property controls this. """
        self._set_iio_dev_attr_str("rx_en", int(value), self._ctrl)

    @property
    def rx_lna_bias_current(self):
        """ Get/Set Rx LNA bias current setting """
        return self._get_iio_dev_attr("bias_current_rx_lna", self._ctrl)

    @rx_lna_bias_current.setter
    def rx_lna_bias_current(self, value):
        """ Get/Set Rx LNA bias current setting """
        self._set_iio_dev_attr_str("bias_current_rx_lna", value, self._ctrl)

    @property
    def rx_lna_enable(self):
        """ Get/Set Rx LNA enable status """
        return bool(self._get_iio_dev_attr("rx_lna_enable", self._ctrl))

    @rx_lna_enable.setter
    def rx_lna_enable(self, value):
        """ Get/Set Rx LNA enable status """
        self._set_iio_dev_attr_str("rx_lna_enable", int(value), self._ctrl)

    @property
    def rx_sequencer_start(self):
        """ Get/Set the Rx Sequencer's starting position """
        return self._get_iio_attr("voltage0", "sequence_start", False)

    @rx_sequencer_start.setter
    def rx_sequencer_start(self, value):
        """ Get/Set the Rx Sequencer's starting position """
        self._set_iio_attr("voltage0", "sequence_start", False, value)

    @property
    def rx_sequencer_stop(self):
        """ Get/Set the Rx Sequencer's ending position """
        return self._get_iio_attr("voltage0", "sequence_end", False)

    @rx_sequencer_stop.setter
    def rx_sequencer_stop(self, value):
        """ Get/Set the Rx Sequencer's ending position """
        self._set_iio_attr("voltage0", "sequence_end", False, value)

    @property
    def rx_to_tx_delay_1(self):
        """ Get/Set Rx to Tx Delay 1 """
        return self._get_iio_dev_attr("rx_to_tx_delay_1", self._ctrl)

    @rx_to_tx_delay_1.setter
    def rx_to_tx_delay_1(self, value):
        """ Get/Set Rx to Tx Delay 1 """
        self._set_iio_dev_attr_str("rx_to_tx_delay_1", int(value), self._ctrl)

    @property
    def rx_to_tx_delay_2(self):
        """ Get/Set Rx to Tx Delay 2 """
        return self._get_iio_dev_attr("rx_to_tx_delay_2", self._ctrl)

    @rx_to_tx_delay_2.setter
    def rx_to_tx_delay_2(self, value):
        """ Get/Set Rx to Tx Delay 2 """
        self._set_iio_dev_attr_str("rx_to_tx_delay_2", int(value), self._ctrl)

    @property
    def rx_vga_enable(self):
        """ Get/Set Rx VGA enable status """
        return bool(self._get_iio_dev_attr("rx_vga_enable", self._ctrl))

    @rx_vga_enable.setter
    def rx_vga_enable(self, value):
        """ Get/Set Rx VGA enable status """
        self._set_iio_dev_attr_str("rx_vga_enable", int(value), self._ctrl)

    @property
    def rx_vga_vm_bias_current(self):
        """ Get/Set Rx VGA/Vector Modulator bias current setting """
        return self._get_iio_dev_attr("bias_current_rx", self._ctrl)

    @rx_vga_vm_bias_current.setter
    def rx_vga_vm_bias_current(self, value):
        """ Get/Set Rx VGA/Vector Modulator bias current setting """
        self._set_iio_dev_attr_str("bias_current_rx", value, self._ctrl)

    @property
    def rx_vm_enable(self):
        """ Get/Set Rx Vector Modulator enable status """
        return bool(self._get_iio_dev_attr("rx_vm_enable", self._ctrl))

    @rx_vm_enable.setter
    def rx_vm_enable(self, value):
        """ Get/Set Rx Vector Modulator enable status """
        self._set_iio_dev_attr_str("rx_vm_enable", int(value), self._ctrl)

    @property
    def sequencer_enable(self):
        """ Get/Set sequencer enable status """
        return bool(self._get_iio_dev_attr("sequencer_enable", self._ctrl))

    @sequencer_enable.setter
    def sequencer_enable(self, value):
        """ Get/Set sequencer enable status """
        self._set_iio_dev_attr_str("sequencer_enable", int(value), self._ctrl)

    @property
    def temperature(self):
        """ Get the temperature reading from the device """
        return self._get_iio_attr("temp0", "raw", False)

    @property
    def _tr(self):
        """Get/Set the status of the T/R input to the device. Valid options are "tx" or "rx". This property must be
        overwritten to use the external T/R input, as GPIO control is not included in this wrapper. This property
        should not be used directly, as it is controlled by the "mode" property.
        """

        if self.tr_source == "external":
            raise NotImplementedError(
                "External T/R control is not implemented in this wrapper. Overwrite this "
                "property in a subclass to allow for GPIO control of the ADAR1000's T/R input pin"
            )

        else:
            return self.tr_spi

    @_tr.setter
    def _tr(self, value):
        """Get/Set the status of the T/R input to the device. Valid options are "tx" or "rx". This property must be
        overwritten to use the external T/R input, as GPIO control is not included in this wrapper. This property
        should not be used directly, as it is controlled by the "mode" property.
        """

        if self.tr_source == "external":
            raise NotImplementedError(
                "External T/R control is not implemented in this wrapper. Overwrite this "
                "property in a subclass to allow for GPIO control of the ADAR1000's T/R input pin"
            )

        else:
            self.tr_spi = value

    @property
    def tr_source(self):
        """ Get/Set TR source for the chip. Valid options are "external" or "spi" """
        value = bool(self._get_iio_dev_attr("tr_source", self._ctrl))
        if value:
            return "external"
        else:
            return "spi"

    @tr_source.setter
    def tr_source(self, value):
        """ Get/Set TR source for the chip. Valid options are "external" or "spi" """
        if value.lower() == "external":
            set_value = 1
        elif value.lower() == "spi":
            set_value = 0
        else:
            raise ValueError('tr_source should be "external" or "spi"')

        self._set_iio_dev_attr_str("tr_source", set_value, self._ctrl)

    @property
    def tr_spi(self):
        """ Get/Set T/R state using the SPI bit. Valid options are "tx" or "rx" """
        value = bool(self._get_iio_dev_attr("tr_spi", self._ctrl))
        if value:
            return "tx"
        else:
            return "rx"

    @tr_spi.setter
    def tr_spi(self, value):
        """ Get/Set T/R state using the SPI bit. Valid options are "tx" or "rx" """
        if value.lower() == "tx":
            set_value = 1
        elif value.lower() == "rx":
            set_value = 0
        else:
            raise ValueError('tr_spi should be "tx" or "rx"')

        self._set_iio_dev_attr_str("tr_spi", set_value, self._ctrl)

    @property
    def tr_switch_enable(self):
        """ Get/Set T/R switch driver enable state """
        return bool(self._get_iio_dev_attr("sw_drv_en_tr", self._ctrl))

    @tr_switch_enable.setter
    def tr_switch_enable(self, value):
        """ Get/Set T/R switch driver enable state """
        self._set_iio_dev_attr_str("sw_drv_en_tr", int(value), self._ctrl)

    @property
    def tx_bias_state(self):
        """ Get/Set the Tx bias memory position when loading from RAM. Valid states are 1-7. """
        return self._get_iio_attr("voltage0", "bias_set_load", True, self._ctrl)

    @tx_bias_state.setter
    def tx_bias_state(self, value):
        """ Get/Set the Tx bias memory position when loading from RAM. Valid states are 1-7. """
        self._set_iio_attr("voltage0", "bias_set_load", True, value, self._ctrl)

    @property
    def _tx_enable(self):
        """ Get/Set enable for entire Tx chain. Should not be used directly, the mode property controls this. """
        return bool(self._get_iio_dev_attr("tx_en", self._ctrl))

    @_tx_enable.setter
    def _tx_enable(self, value):
        """ Get/Set enable for entire Tx chain. Should not be used directly, the mode property controls this. """
        self._set_iio_dev_attr_str("tx_en", int(value), self._ctrl)

    @property
    def tx_pa_bias_current(self):
        """ Get/Set Tx PA bias current setting """
        return self._get_iio_dev_attr("bias_current_tx_drv", self._ctrl)

    @tx_pa_bias_current.setter
    def tx_pa_bias_current(self, value):
        """ Get/Set Tx PA bias current setting """
        self._set_iio_dev_attr_str("bias_current_tx_drv", value, self._ctrl)

    @property
    def tx_pa_enable(self):
        """ Get/Set Tx PA enable status """
        return bool(self._get_iio_dev_attr("tx_drv_enable", self._ctrl))

    @tx_pa_enable.setter
    def tx_pa_enable(self, value):
        """ Get/Set Tx PA enable status """
        self._set_iio_dev_attr_str("tx_drv_enable", int(value), self._ctrl)

    @property
    def tx_sequencer_start(self):
        """ Get/Set the Tx Sequencer's starting position """
        return self._get_iio_attr("voltage0", "sequence_start", True)

    @tx_sequencer_start.setter
    def tx_sequencer_start(self, value):
        """ Get/Set the Tx Sequencer's starting position """
        self._set_iio_attr("voltage0", "sequence_start", True, value)

    @property
    def tx_sequencer_stop(self):
        """ Get/Set the Tx Sequencer's ending position """
        return self._get_iio_attr("voltage0", "sequence_end", True)

    @tx_sequencer_stop.setter
    def tx_sequencer_stop(self, value):
        """ Get/Set the Tx Sequencer's ending position """
        self._set_iio_attr("voltage0", "sequence_end", True, value)

    @property
    def tx_to_rx_delay_1(self):
        """ Get/Set Tx to Rx Delay 1 """
        return self._get_iio_dev_attr("tx_to_rx_delay_1", self._ctrl)

    @tx_to_rx_delay_1.setter
    def tx_to_rx_delay_1(self, value):
        """ Get/Set Tx to Rx Delay 1 """
        self._set_iio_dev_attr_str("tx_to_rx_delay_1", int(value), self._ctrl)

    @property
    def tx_to_rx_delay_2(self):
        """ Get/Set Tx to Rx Delay 2 """
        return self._get_iio_dev_attr("tx_to_rx_delay_2", self._ctrl)

    @tx_to_rx_delay_2.setter
    def tx_to_rx_delay_2(self, value):
        """ Get/Set Tx to Rx Delay 2 """
        self._set_iio_dev_attr_str("tx_to_rx_delay_2", int(value), self._ctrl)

    @property
    def tx_vga_enable(self):
        """ Get/Set Tx VGA enable status """
        return bool(self._get_iio_dev_attr("tx_vga_enable", self._ctrl))

    @tx_vga_enable.setter
    def tx_vga_enable(self, value):
        """ Get/Set Tx VGA enable status """
        self._set_iio_dev_attr_str("tx_vga_enable", int(value), self._ctrl)

    @property
    def tx_vga_vm_bias_current(self):
        """ Get/Set Tx VGA/Vector Modulator bias current setting """
        return self._get_iio_dev_attr("bias_current_tx", self._ctrl)

    @tx_vga_vm_bias_current.setter
    def tx_vga_vm_bias_current(self, value):
        """ Get/Set Tx VGA/Vector Modulator bias current setting """
        self._set_iio_dev_attr_str("bias_current_tx", value, self._ctrl)

    @property
    def tx_vm_enable(self):
        """ Get/Set Tx Vector Modulator enable status """
        return bool(self._get_iio_dev_attr("tx_vm_enable", self._ctrl))

    @tx_vm_enable.setter
    def tx_vm_enable(self, value):
        """ Get/Set Tx Vector Modulator enable status """
        self._set_iio_dev_attr_str("tx_vm_enable", int(value), self._ctrl)

    """ Public Methods """

    def generate_clocks(self):
        """ Generate CLK cycles before pulsing RX_LOAD or TX_LOAD """
        self._set_iio_dev_attr_str("gen_clk_cycles", "", self._ctrl)

    def initialize(self, pa_off=-2.5, pa_on=-2.5, lna_off=-2, lna_on=-2):
        """Suggested initialization routine after powerup

        parameters:
            pa_off: float
                Voltage to set the PA_BIAS_OFF values to during initialization
            pa_on: float
                Voltage to set the PA_BIAS_ON values to during initialization
            lna_off: float
                Voltage to set the LNA_BIAS_OFF values to during initialization
            lna_on: float
                Voltage to set the LNA_BIAS_ON values to during initialization
        """
        # Put the part in a known state
        self.reset()

        # Set the bias currents to nominal
        self.rx_lna_bias_current = 0x08
        self.rx_vga_vm_bias_current = 0x55
        self.tx_vga_vm_bias_current = 0x2D
        self.tx_pa_bias_current = 0x06

        # Disable RAM control
        self.beam_mem_enable = False
        self.bias_mem_enable = False
        self.common_mem_enable = False

        # Enable all internal amplifiers
        self.rx_vga_enable = True
        self.rx_vm_enable = True
        self.rx_lna_enable = True
        self.tx_vga_enable = True
        self.tx_vm_enable = True
        self.tx_pa_enable = True

        # Disable Tx/Rx paths for the device
        self.mode = "disabled"

        # Enable the PA/LNA bias DACs
        self.lna_bias_out_enable = True
        self.bias_dac_enable = True
        self.bias_dac_mode = "toggle"

        # Configure the external switch control
        self.external_tr_polarity = True
        self.tr_switch_enable = True

        # Set the default LNA bias
        self.lna_bias_off = lna_off
        self.lna_bias_on = lna_on

        # Settings for each channel
        for channel in self.channels:
            # Default channel enable
            channel.rx_enable = False
            channel.tx_enable = False

            # Default PA bias
            channel.pa_bias_off = pa_off
            channel.pa_bias_on = pa_on

            # Default attenuator, gain, and phase
            channel.rx_attenuator = False
            channel.rx_gain = 0x7F
            channel.rx_phase = 0
            channel.tx_attenuator = False
            channel.tx_gain = 0x7F
            channel.tx_phase = 0

        # Latch in the new settings
        self.latch_rx_settings()
        self.latch_tx_settings()

    def latch_rx_settings(self):
        """ Latch in new Gain/Phase settings for the Rx """
        self._set_iio_dev_attr_str("rx_load_spi", 1, self._ctrl)

    def latch_tx_settings(self):
        """ Latch in new Gain/Phase settings for the Tx """
        self._set_iio_dev_attr_str("tx_load_spi", 1, self._ctrl)

    def reset(self):
        """ Reset ADAR1000 to default settings """
        self._set_iio_dev_attr("reset", 1, self._ctrl)

    def save_rx_bias(
        self,
        state,
        lna_bias_off,
        lna_bias_on,
        rx_vga_vm_bias_current,
        rx_lna_bias_current,
    ):
        """Save a bias setting to an Rx memory position

        parameters:
            state: int
                State number to save. Valid options are 1 to 7
            lna_bias_off: float
                LNA_BIAS_OFF voltage
            lna_bias_on: float
                LNA_BIAS_ON voltage
            rx_vga_vm_bias_current: int
                Bias current setting for the Rx VGA and Vector Modulator
            rx_lna_bias_current: int
                Bias current setting for the Rx LNA
        """

        # Convert the LNA bias settings to dac codes:
        lna_off_dac_code = int(lna_bias_off / self._BIAS_CODE_TO_VOLTAGE_SCALE)
        lna_on_dac_code = int(lna_bias_on / self._BIAS_CODE_TO_VOLTAGE_SCALE)

        save_string = f"{state}, {lna_off_dac_code}, {lna_on_dac_code}, {rx_vga_vm_bias_current}, {rx_lna_bias_current}"

        self._set_iio_attr("voltage0", "bias_set_save", False, save_string, self._ctrl)

    def save_tx_bias(
        self,
        state,
        pa1_bias_off,
        pa2_bias_off,
        pa3_bias_off,
        pa4_bias_off,
        pa1_bias_on,
        pa2_bias_on,
        pa3_bias_on,
        pa4_bias_on,
        tx_vga_vm_bias_current,
        tx_pa_bias_current,
    ):
        """Save a bias setting to a Tx memory position

        parameters:
            state: int
                State number to save. Valid options are 1 to 7
            pa1_bias_off: float
                PA1_BIAS_OFF voltage
            pa2_bias_off: float
                PA2_BIAS_OFF voltage
            pa3_bias_off: float
                PA3_BIAS_OFF voltage
            pa4_bias_off: float
                PA4_BIAS_OFF voltage
            pa1_bias_on: float
                PA1_BIAS_ON voltage
            pa2_bias_on: float
                PA2_BIAS_ON voltage
            pa3_bias_on: float
                PA3_BIAS_ON voltage
            pa4_bias_on: float
                PA4_BIAS_ON voltage
            tx_vga_vm_bias_current: int
                Bias current setting for the Tx VGA and Vector Modulator
            tx_lna_bias_current: int
                Bias current setting for the Tx PA
        """

        # Convert the PA bias settings to dac codes:
        pa1_off_dac_code = int(pa1_bias_off / self._BIAS_CODE_TO_VOLTAGE_SCALE)
        pa2_off_dac_code = int(pa2_bias_off / self._BIAS_CODE_TO_VOLTAGE_SCALE)
        pa3_off_dac_code = int(pa3_bias_off / self._BIAS_CODE_TO_VOLTAGE_SCALE)
        pa4_off_dac_code = int(pa4_bias_off / self._BIAS_CODE_TO_VOLTAGE_SCALE)
        pa1_on_dac_code = int(pa1_bias_on / self._BIAS_CODE_TO_VOLTAGE_SCALE)
        pa2_on_dac_code = int(pa2_bias_on / self._BIAS_CODE_TO_VOLTAGE_SCALE)
        pa3_on_dac_code = int(pa3_bias_on / self._BIAS_CODE_TO_VOLTAGE_SCALE)
        pa4_on_dac_code = int(pa4_bias_on / self._BIAS_CODE_TO_VOLTAGE_SCALE)

        save_string = (
            f"{state}, {pa1_off_dac_code}, {pa2_off_dac_code}, {pa3_off_dac_code}, {pa4_off_dac_code}, "
            f"{pa1_on_dac_code}, {pa2_on_dac_code}, {pa3_on_dac_code}, {pa4_on_dac_code}, {tx_vga_vm_bias_current}, "
            f"{tx_pa_bias_current}"
        )

        self._set_iio_attr("voltage0", "bias_set_save", True, save_string, self._ctrl)


class adar1000_array(context_manager):
    """ADAR1000 Beamformer Array

    parameters:
        uri: type=string
            URI of IIO context with ADAR1000 array
        chip_ids: type=list[string]
            List of strings identifying desired chip select and hardware ID
            for the ADAR1000. These strings are the labels coinciding with
            each chip select and hardware address and are typically in the
            form csbX_chipX. The csb line can be any number depending on how
            many are used in the system. The chip number will typically be
            1-4 because each CSB line can control up to four ADAR1000s. Note
            that the order of the devices listed will correspond to the
            device numbers in the array map directly.
        device_map: type=list[list[int]]
            List with the map of where the ADAR1000s are in the array. Each
            entry in the map represents a row of ADAR1000s referenced by
            device number. For example, a map:

                | [[1, 3, 5, 7],
                | [2, 4, 6, 8]]

            represents an array of 8 ADAR1000s 4 wide and 2 tall.
        element_map: type=list[list[int]]
            List of lists with the map of where the array elements are in the
            physical array. Each entry in the map represents a row of array
            channels referenced by element number. For example, a map:

                | [[1, 5, 9, 13],
                | [2, 6, 10, 14],
                | [3, 7, 11, 15],
                | [4, 8, 12, 16]]

            represents an array of 16 elements (4 ADAR1000s) in a square array.
        device_element_map: type=dict[int, list[int]]
            Dictionary with the map of ADAR1000 to array element references. Each
            key in the map is a device number. The corresponding list of integers
            represents the array element numbers connected to that ADAR1000, in
            order of the ADAR1000's channels. For example, an entry of
            {3: [10, 14, 13, 9]} connects ADAR1000 #3 to array elements 10, 14, 13,
            and 9. Element #10 is on the ADAR1000's channel 1 while element #13 is
            on the ADAR1000's channel 3.
    """

    _device_name = ""

    def __init__(
        self,
        uri="",
        chip_ids=None,
        device_map=None,
        element_map=None,
        device_element_map=None,
    ):
        # Check that the number of chips matches for all the inputs
        chip_ids_len = len(chip_ids)
        adar1000_map_len = len([num for row in device_map for num in row])
        element_map_len = len([num for row in element_map for num in row]) / 4
        device_element_map_len = len(device_element_map)
        if not (
            chip_ids_len
            == adar1000_map_len
            == element_map_len
            == device_element_map_len
        ):
            raise ValueError("The number of chips must match for all the inputs")

        # Get the context
        context_manager.__init__(self, uri, self._device_name)

        # Initialize the devices dictionary
        self._devices = {}

        # Get an ADAR1000 handle for each chip ID using the context already found
        for i, chip_id in enumerate(chip_ids):
            device_number = i + 1
            self._devices[chip_id] = adar1000(
                uri,
                self._ctx,
                chip_id,
                device_number,
                element_map,
                device_element_map[device_number],
            )

        if len(self.devices) != len(chip_ids):
            raise Exception(
                f"Couldn't find all devices in the array {', '.join(chip_ids)}"
            )

        # Initialize some class fields
        self._chip_ids = chip_ids
        self._device_map = device_map
        self._element_map = element_map
        self._element_spacing = 0.015
        self._frequency = 10e9
        self._rx_azimuth = 0
        self._rx_azimuth_phi = 0
        self._rx_elevation = 0
        self._rx_elevation_phi = 0
        self._tx_azimuth = 0
        self._tx_azimuth_phi = 0
        self._tx_elevation = 0
        self._tx_elevation_phi = 0

    def __repr__(self):
        """ Representation of the ADAR1000 array class """
        return f"ADAR1000 array containing {len(self.devices)} ADAR1000 instances:\n\t{', '.join(self._chip_ids)}"

    """ Array Properties """

    @property
    def all_rx_attenuators(self):
        """Get/Set all Rx Attenuator settings in the array.
        The format is a list of lists where each row in the
        array is a list entry in the larger list."""
        attenuator_map = []
        for r, row in enumerate(self.element_map):
            attenuator_map.append([])
            for element in row:
                attenuator_map[r].append(self.elements[element].rx_attenuator)

        return attenuator_map

    @all_rx_attenuators.setter
    def all_rx_attenuators(self, value):
        """Get/Set all Rx Attenuator settings in the array.
        The format is a list of lists where each row in the
        array is a list entry in the larger list."""
        for r, row in enumerate(self.element_map):
            for e, element in enumerate(row):
                self.elements[element].rx_attenuator = value[r][e]

    @property
    def all_rx_gains(self):
        """Get/Set all Rx Gain settings in the array.
        The format is a list of lists where each row in the
        array is a list entry in the larger list."""
        gain_map = []
        for r, row in enumerate(self.element_map):
            gain_map.append([])
            for element in row:
                gain_map[r].append(self.elements[element].rx_gain)

        return gain_map

    @all_rx_gains.setter
    def all_rx_gains(self, value):
        """Get/Set all Rx Gain settings in the array.
        The format is a list of lists where each row in the
        array is a list entry in the larger list."""
        for r, row in enumerate(self.element_map):
            for e, element in enumerate(row):
                self.elements[element].rx_gain = value[r][e]

    @property
    def all_rx_phases(self):
        """Get/Set all Rx Phase settings in the array.
        The format is a list of lists where each row in the
        array is a list entry in the larger list."""
        phase_map = []
        for r, row in enumerate(self.element_map):
            phase_map.append([])
            for element in row:
                phase_map[r].append(self.elements[element].rx_phase)

        return phase_map

    @all_rx_phases.setter
    def all_rx_phases(self, value):
        """Get/Set all Rx Phase settings in the array.
        The format is a list of lists where each row in the
        array is a list entry in the larger list."""
        for r, row in enumerate(self.element_map):
            for e, element in enumerate(row):
                self.elements[element].rx_phase = value[r][e]

    @property
    def all_tx_attenuators(self):
        """Get/Set all Tx Attenuator settings in the array.
        The format is a list of lists where each row in the
        array is a list entry in the larger list."""
        attenuator_map = []
        for r, row in enumerate(self.element_map):
            attenuator_map.append([])
            for element in row:
                attenuator_map[r].append(self.elements[element].tx_attenuator)

        return attenuator_map

    @all_tx_attenuators.setter
    def all_tx_attenuators(self, value):
        """Get/Set all Tx Attenuator settings in the array.
        The format is a list of lists where each row in the
        array is a list entry in the larger list."""
        for r, row in enumerate(self.element_map):
            for e, element in enumerate(row):
                self.elements[element].tx_attenuator = value[r][e]

    @property
    def all_tx_gains(self):
        """Get/Set all Tx Gain settings in the array.
        The format is a list of lists where each row in the
        array is a list entry in the larger list."""
        gain_map = []
        for r, row in enumerate(self.element_map):
            gain_map.append([])
            for element in row:
                gain_map[r].append(self.elements[element].tx_gain)

        return gain_map

    @all_tx_gains.setter
    def all_tx_gains(self, value):
        """Get/Set all Tx Gain settings in the array.
        The format is a list of lists where each row in the
        array is a list entry in the larger list."""
        for r, row in enumerate(self.element_map):
            for e, element in enumerate(row):
                self.elements[element].tx_gain = value[r][e]

    @property
    def all_tx_phases(self):
        """Get/Set all Tx Phase settings in the array.
        The format is a list of lists where each row in the
        array is a list entry in the larger list."""
        phase_map = []
        for r, row in enumerate(self.element_map):
            phase_map.append([])
            for element in row:
                phase_map[r].append(self.elements[element].tx_phase)

        return phase_map

    @all_tx_phases.setter
    def all_tx_phases(self, value):
        """Get/Set all Tx Phase settings in the array.
        The format is a list of lists where each row in the
        array is a list entry in the larger list."""
        for r, row in enumerate(self.element_map):
            for e, element in enumerate(row):
                self.elements[element].tx_phase = value[r][e]

    @property
    def devices(self):
        """
        Dictionary representing all of the connected ADAR1000s.
        The dictionary key is the chip_id for each device.
        """
        return {device.array_device_number: device for device in self._devices.values()}

    @property
    def device_map(self):
        """ Get the map of ADAR1000s in the array """
        return self._device_map

    @property
    def elements(self):
        """
        Dictionary representing all of the connected ADAR1000 elements sorted by element number.
        The dictionary key is the element number for each device.
        """
        dictionary = {
            el.array_element_number: el
            for dev in self.devices.values()
            for el in dev.channels
        }
        return {k: v for k, v in sorted(dictionary.items(), key=lambda x: x[0])}

    @property
    def element_map(self):
        """ Get the map of elements in the array """
        return self._element_map

    @property
    def element_spacing(self):
        """ Get/Set the element spacing for the array in meters """
        return self._element_spacing

    @element_spacing.setter
    def element_spacing(self, value):
        """ Get/Set the element spacing for the array in meters """
        self._element_spacing = value

    @property
    def frequency(self):
        """ Get/Set the board frequency in Hz """
        return self._frequency

    @frequency.setter
    def frequency(self, value):
        """ Get/Set the board frequency in Hz """
        self._frequency = value

    @property
    def rx_azimuth(self):
        """ Get the Rx azimuth angle for the array in degrees """
        return self._rx_azimuth

    @property
    def rx_azimuth_phi(self):
        """ Get the Rx azimuth phi angle for the array in degrees """
        return self._rx_azimuth_phi

    @property
    def rx_elevation(self):
        """ Get the Rx elevation angle for the array in degrees """
        return self._rx_elevation

    @property
    def rx_elevation_phi(self):
        """ Get the Rx elevation phi angle for the array in degrees """
        return self._rx_elevation_phi

    @property
    def temperatures(self):
        """ Get the temperature readings of the ADAR1000s in a dictionary """
        return {chip_id: device.temperature for chip_id, device in self.devices.items()}

    @property
    def tx_azimuth(self):
        """ Get the Tx azimuth angle for the array in degrees """
        return self._tx_azimuth

    @property
    def tx_azimuth_phi(self):
        """ Get the Tx azimuth phi angle for the array in degrees """
        return self._tx_azimuth_phi

    @property
    def tx_elevation(self):
        """ Get the Tx elevation angle for the array in degrees """
        return self._tx_elevation

    @property
    def tx_elevation_phi(self):
        """ Get the Tx elevation phi angle for the array in degrees """
        return self._tx_elevation_phi

    def _steer(self, rx_or_tx, azimuth, elevation):
        """Steer the array
        parameters:
            rx_or_tx: string
                Sets which parameters are updated, Rx or Tx.
            azimuth: float
                Desired beam angle in degrees for the horizontal direction
            elevation: float
                Desired beam angle in degrees for the vertical direction
        """

        # Clean up the rx_or_tx variable
        rx_or_tx = rx_or_tx.strip().lower()

        # Calculate the Φ angles for each element in both directions (in degrees)
        azimuth_phi, elevation_phi = self.calculate_phi(azimuth, elevation)

        # Update the class variables
        if rx_or_tx == "rx":
            self._rx_azimuth = azimuth
            self._rx_elevation = elevation
            self._rx_azimuth_phi = azimuth_phi
            self._rx_elevation_phi = elevation_phi
        else:
            self._tx_azimuth = azimuth
            self._tx_elevation = elevation
            self._tx_azimuth_phi = azimuth_phi
            self._tx_elevation_phi = elevation_phi

        # Steer the elements in the array
        for element in self.elements.values():
            # Calculate the row and column phases
            column_phase = element.column * azimuth_phi
            row_phase = element.row * elevation_phi

            if rx_or_tx == "rx":
                element.rx_phase = column_phase + row_phase
            else:
                element.tx_phase = column_phase + row_phase

        # Latch in the new phases
        if rx_or_tx == "rx":
            self.latch_rx_settings()
        else:
            self.latch_tx_settings()

    """ Public Methods """

    def calculate_phi(self, azimuth, elevation):
        """Calculate the Φ angles to steer the array in a particular direction. This method assumes that the entire
            array is one analog beam.
        parameters:
            azimuth: float
                Desired beam angle in degrees for the horizontal direction.
            elevation: float
                Desired beam angle in degrees for the vertical direction.
        """

        # Convert the input angles to radians
        az_rads = azimuth * pi / 180
        el_rads = elevation * pi / 180

        # Calculate the phase increment (Φ) for each element in the array in both directions (in degrees)
        az_phi = 2 * self.frequency * self.element_spacing * sin(az_rads) * 180 / 3e8
        el_phi = 2 * self.frequency * self.element_spacing * sin(el_rads) * 180 / 3e8

        return az_phi, el_phi

    def initialize_devices(self, pa_off=-2.5, pa_on=-2.5, lna_off=-2, lna_on=-2):
        """Suggested initialization routine after powerup

        parameters:
            pa_off: float
                Voltage to set the PA_BIAS_OFF values to during initialization
            pa_on: float
                Voltage to set the PA_BIAS_ON values to during initialization
            lna_off: float
                Voltage to set the LNA_BIAS_OFF values to during initialization
            lna_on: float
                Voltage to set the LNA_BIAS_ON values to during initialization
        """
        for device in self.devices.values():
            device.initialize(
                pa_off=pa_off, pa_on=pa_on, lna_off=lna_off, lna_on=lna_on
            )

    def latch_rx_settings(self):
        """ Latch in new Gain/Phase settings for the Rx """
        for device in self.devices.values():
            device.latch_rx_settings()

    def latch_tx_settings(self):
        """ Latch in new Gain/Phase settings for the Tx """
        for device in self.devices.values():
            device.latch_tx_settings()

    def steer_rx(self, azimuth, elevation):
        """Steer the Rx array in a particular direction. This method assumes that the entire array is one analog beam.

        parameters:
            azimuth: float
                Desired beam angle in degrees for the horizontal direction.
            elevation: float
                Desired beam angle in degrees for the vertical direction.
        """

        self._steer("rx", azimuth, elevation)

    def steer_tx(self, azimuth, elevation):
        """Steer the Tx array in a particular direction. This method assumes that the entire array is one analog beam.

        parameters:
            azimuth: float
                Desired beam angle in degrees for the horizontal direction.
            elevation: float
                Desired beam angle in degrees for the vertical direction.
        """

        self._steer("tx", azimuth, elevation)
